@@ -1,5 +1,5 @@
 /* Buffer management for tar.
-   Copyright (C) 1988, 92, 93, 94, 96, 97, 1999 Free Software Foundation, Inc.
+   Copyright 1988, 92, 93, 94, 96, 97, 99, 2000 Free Software Foundation, Inc.
    Written by John Gilmore, on 1985-08-25.
 
    This program is free software; you can redistribute it and/or modify it
@@ -16,11 +16,18 @@
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
+/* Enable GNU extensions in fnmatch.h.  */
+#ifndef _GNU_SOURCE
+# define _GNU_SOURCE 1
+#endif
+
 #include "system.h"
 
 #include <signal.h>
 #include <time.h>
+#ifndef time
 time_t time ();
+#endif
 
 #if MSDOS
 # include <process.h>
@@ -30,15 +37,13 @@ time_t time ();
 # include <sys/inode.h>
 #endif
 
-#ifndef FNM_LEADING_DIR
-# include <fnmatch.h>
-#endif
+#include <fnmatch.h>
+#include <human.h>
+#include <quotearg.h>
 
 #include "common.h"
 #include "rmt.h"
 
-#define DEBUG_FORK 0		/* if nonzero, childs are born stopped */
-
 #define	PREAD 0			/* read file descriptor from pipe() */
 #define	PWRITE 1		/* write file descriptor from pipe() */
 
@@ -50,7 +55,7 @@ time_t time ();
 
 /* Variables.  */
 
-static tarlong total_written;	/* bytes written on all volumes */
+static tarlong prev_written;	/* bytes written on previous volumes */
 static tarlong bytes_written;	/* bytes written on this volume */
 
 /* FIXME: The following four variables should ideally be static to this
@@ -71,8 +76,8 @@ FILE *stdlis;
 
 static void backspace_output PARAMS ((void));
 static int new_volume PARAMS ((enum access_mode));
-static void write_error PARAMS ((ssize_t));
-static void read_error PARAMS ((void));
+static void archive_write_error PARAMS ((ssize_t)) __attribute__ ((noreturn));
+static void archive_read_error PARAMS ((void));
 
 #if !MSDOS
 /* Obnoxious test to see if dimwit is trying to dump the archive.  */
@@ -114,49 +119,46 @@ static int global_volno = 1;	/* volume number to print in external
 
 char *save_name;		/* name of the file we are currently writing */
 off_t save_totsize;		/* total size of file we are writing, only
-				   valid if save_name is non NULL */
+				   valid if save_name is nonzero */
 off_t save_sizeleft;		/* where we are in the file we are writing,
 				   only valid if save_name is nonzero */
 
-int write_archive_to_stdout = 0;
+int write_archive_to_stdout;
 
 /* Used by flush_read and flush_write to store the real info about saved
    names.  */
-static char *real_s_name = NULL;
+static char *real_s_name;
 static off_t real_s_totsize;
 static off_t real_s_sizeleft;
 
 /* Functions.  */
 
-#if DEBUG_FORK
-
-static pid_t
-myfork (void)
-{
-  pid_t result = fork();
-
-  if (result == 0)
-    kill (getpid (), SIGSTOP);
-  return result;
-}
-
-# define fork myfork
-
-#endif /* DEBUG FORK */
-
-void
-init_total_written (void)
-{
-  clear_tarlong (total_written);
-  clear_tarlong (bytes_written);
-}
-
 void
 print_total_written (void)
 {
-  fprintf (stderr, _("Total bytes written: "));
-  print_tarlong (total_written, stderr);
-  fprintf (stderr, "\n");
+  tarlong written = prev_written + bytes_written;
+  char bytes[sizeof (tarlong) * CHAR_BIT];
+  char abbr[LONGEST_HUMAN_READABLE + 1];
+  char rate[LONGEST_HUMAN_READABLE + 1];
+  double seconds;
+
+#if HAVE_CLOCK_GETTIME
+  struct timespec now;
+  if (clock_gettime (CLOCK_REALTIME, &now) == 0)
+    seconds = ((now.tv_sec - start_timespec.tv_sec)
+	       + (now.tv_nsec - start_timespec.tv_nsec) / 1e9);
+  else
+#endif
+    seconds = time (0) - start_time;
+
+  sprintf (bytes, TARLONG_FORMAT, written);
+
+  /* Amanda 2.4.1p1 looks for "Total bytes written: [0-9][0-9]*".  */
+  fprintf (stderr, _("Total bytes written: %s (%sB, %sB/s)\n"), bytes,
+	   human_readable ((uintmax_t) written, abbr, 1, -1024),
+	   (0 < seconds && written / seconds < (uintmax_t) -1
+	    ? human_readable ((uintmax_t) (written / seconds), rate, 1, -1024)
+	    : "?"));
 }
 
 /*--------------------------------------------------------.
@@ -187,7 +189,7 @@ reset_eof (void)
 
 /*-------------------------------------------------------------------------.
 | Return the location of the next available input or output block.	   |
-| Return NULL for EOF.  Once we have returned NULL, we just keep returning |
+| Return zero for EOF.  Once we have returned zero, we just keep returning |
 | it, to avoid accidentally going on to the next file on the tape.	   |
 `-------------------------------------------------------------------------*/
 
@@ -197,12 +199,12 @@ find_next_block (void)
   if (current_block == record_end)
     {
       if (hit_eof)
-	return NULL;
+	return 0;
       flush_archive ();
       if (current_block == record_end)
 	{
 	  hit_eof = 1;
-	  return NULL;
+	  return 0;
 	}
     }
   return current_block;
@@ -242,33 +244,41 @@ available_space_after (union block *pointer)
 }
 
 /*------------------------------------------------------------------.
-| Close file having descriptor FD, and abort if close unsucessful.  |
+| Close file having descriptor FD, and abort if close unsuccessful. |
 `------------------------------------------------------------------*/
 
 static void
 xclose (int fd)
 {
-  if (close (fd) < 0)
-    FATAL_ERROR ((0, errno, _("Cannot close file #%d"), fd));
+  if (close (fd) != 0)
+    close_error (_("(pipe)"));
 }
 
-/*-----------------------------------------------------------------------.
-| Duplicate file descriptor FROM into becoming INTO, or else, issue	 |
-| MESSAGE.  INTO is closed first and has to be the next available slot.	 |
-`-----------------------------------------------------------------------*/
+/* Duplicate file descriptor FROM into becoming INTO.
+   INTO is closed first and has to be the next available slot.  */
 
 static void
-xdup2 (int from, int into, const char *message)
+xdup2 (int from, int into)
 {
   if (from != into)
     {
       int status = close (into);
 
-      if (status < 0 && errno != EBADF)
-	FATAL_ERROR ((0, errno, _("Cannot close descriptor %d"), into));
+      if (status != 0 && errno != EBADF)
+	{
+	  int e = errno;
+	  FATAL_ERROR ((0, e, _("Cannot close")));
+	}
       status = dup (from);
       if (status != into)
-	FATAL_ERROR ((0, errno, _("Cannot properly duplicate %s"), message));
+	{
+	  if (status < 0)
+	    {
+	      int e = errno;
+	      FATAL_ERROR ((0, e, _("Cannot dup")));
+	    }
+	  abort ();
+	}
       xclose (from);
     }
 }
@@ -307,13 +317,10 @@ is_regular_file (const char *name)
 {
   struct stat stbuf;
 
-  if (stat (name, &stbuf) < 0)
-    return 1;
-
-  if (S_ISREG (stbuf.st_mode))
-    return 1;
-
-  return 0;
+  if (stat (name, &stbuf) == 0)
+    return S_ISREG (stbuf.st_mode);
+  else
+    return errno == ENOENT;
 }
 
 static ssize_t
@@ -345,12 +352,8 @@ child_open_for_compress (void)
   int child_pipe[2];
   pid_t grandchild_pid;
 
-  if (pipe (parent_pipe) < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open pipe")));
-
-  child_pid = fork ();
-  if (child_pid < 0)
-    FATAL_ERROR ((0, errno, _("Cannot fork")));
+  xpipe (parent_pipe);
+  child_pid = xfork ();
 
   if (child_pid > 0)
     {
@@ -365,7 +368,7 @@ child_open_for_compress (void)
 
   program_name = _("tar (child)");
 
-  xdup2 (parent_pipe[PREAD], STDIN_FILENO, _("(child) Pipe to stdin"));
+  xdup2 (parent_pipe[PREAD], STDIN_FILENO);
   xclose (parent_pipe[PWRITE]);
 
   /* Check if we need a grandchild tar.  This happens only if either:
@@ -390,36 +393,29 @@ child_open_for_compress (void)
 
 	  if (backup_option)
 	    undo_last_backup ();
-	  FATAL_ERROR ((0, saved_errno, _("Cannot open archive %s"),
-			archive_name_array[0]));
+	  errno = saved_errno;
+	  open_fatal (archive_name_array[0]);
 	}
-      xdup2 (archive, STDOUT_FILENO, _("Archive to stdout"));
+      xdup2 (archive, STDOUT_FILENO);
       execlp (use_compress_program_option, use_compress_program_option,
 	      (char *) 0);
-      FATAL_ERROR ((0, errno, _("Cannot exec %s"),
-		    use_compress_program_option));
+      exec_fatal (use_compress_program_option);
     }
 
   /* We do need a grandchild tar.  */
 
-  if (pipe (child_pipe) < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open pipe")));
-
-  grandchild_pid = fork ();
-  if (grandchild_pid < 0)
-    FATAL_ERROR ((0, errno, _("Child cannot fork")));
+  xpipe (child_pipe);
+  grandchild_pid = xfork ();
 
   if (grandchild_pid > 0)
     {
       /* The child tar is still here!  Launch the compressor.  */
 
-      xdup2 (child_pipe[PWRITE], STDOUT_FILENO,
-	     _("((child)) Pipe to stdout"));
+      xdup2 (child_pipe[PWRITE], STDOUT_FILENO);
       xclose (child_pipe[PREAD]);
       execlp (use_compress_program_option, use_compress_program_option,
 	      (char *) 0);
-      FATAL_ERROR ((0, errno, _("Cannot exec %s"),
-		    use_compress_program_option));
+      exec_fatal (use_compress_program_option);
     }
 
   /* The new born grandchild tar is here!  */
@@ -428,16 +424,17 @@ child_open_for_compress (void)
 
   /* Prepare for reblocking the data from the compressor into the archive.  */
 
-  xdup2 (child_pipe[PREAD], STDIN_FILENO, _("(grandchild) Pipe to stdin"));
+  xdup2 (child_pipe[PREAD], STDIN_FILENO);
   xclose (child_pipe[PWRITE]);
 
   if (strcmp (archive_name_array[0], "-") == 0)
     archive = STDOUT_FILENO;
   else
-    archive = rmtcreat (archive_name_array[0], MODE_RW, rsh_command_option);
-  if (archive < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open archive %s"),
-		  archive_name_array[0]));
+    {
+      archive = rmtcreat (archive_name_array[0], MODE_RW, rsh_command_option);
+      if (archive < 0)
+	open_fatal (archive_name_array[0]);
+    }
 
   /* Let's read out of the stdin pipe and write an archive.  */
 
@@ -463,7 +460,7 @@ child_open_for_compress (void)
 	}
 
       if (status < 0)
-	FATAL_ERROR ((0, errno, _("Cannot read from compression program")));
+	read_fatal (use_compress_program_option);
 
       /* Copy the record.  */
 
@@ -478,7 +475,7 @@ child_open_for_compress (void)
 	      memset (record_start->buffer + length, 0, record_size - length);
 	      status = write_archive_buffer ();
 	      if (status != record_size)
-		write_error (status);
+		archive_write_error (status);
 	    }
 
 	  /* There is nothing else to read, break out.  */
@@ -487,7 +484,7 @@ child_open_for_compress (void)
 
       status = write_archive_buffer ();
       if (status != record_size)
- 	write_error (status);
+ 	archive_write_error (status);
     }
 
 #if 0
@@ -507,12 +504,8 @@ child_open_for_uncompress (void)
   int child_pipe[2];
   pid_t grandchild_pid;
 
-  if (pipe (parent_pipe) < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open pipe")));
-
-  child_pid = fork ();
-  if (child_pid < 0)
-    FATAL_ERROR ((0, errno, _("Cannot fork")));
+  xpipe (parent_pipe);
+  child_pid = xfork ();
 
   if (child_pid > 0)
     {
@@ -528,7 +521,7 @@ child_open_for_uncompress (void)
 
   program_name = _("tar (child)");
 
-  xdup2 (parent_pipe[PWRITE], STDOUT_FILENO, _("(child) Pipe to stdout"));
+  xdup2 (parent_pipe[PWRITE], STDOUT_FILENO);
   xclose (parent_pipe[PREAD]);
 
   /* Check if we need a grandchild tar.  This happens only if either:
@@ -545,34 +538,27 @@ child_open_for_uncompress (void)
 
       archive = open (archive_name_array[0], O_RDONLY | O_BINARY, MODE_RW);
       if (archive < 0)
-	FATAL_ERROR ((0, errno, _("Cannot open archive %s"),
-		      archive_name_array[0]));
-      xdup2 (archive, STDIN_FILENO, _("Archive to stdin"));
+	open_fatal (archive_name_array[0]);
+      xdup2 (archive, STDIN_FILENO);
       execlp (use_compress_program_option, use_compress_program_option,
 	      "-d", (char *) 0);
-      FATAL_ERROR ((0, errno, _("Cannot exec %s"),
-		    use_compress_program_option));
+      exec_fatal (use_compress_program_option);
     }
 
   /* We do need a grandchild tar.  */
 
-  if (pipe (child_pipe) < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open pipe")));
-
-  grandchild_pid = fork ();
-  if (grandchild_pid < 0)
-    FATAL_ERROR ((0, errno, _("Child cannot fork")));
+  xpipe (child_pipe);
+  grandchild_pid = xfork ();
 
   if (grandchild_pid > 0)
     {
       /* The child tar is still here!  Launch the uncompressor.  */
 
-      xdup2 (child_pipe[PREAD], STDIN_FILENO, _("((child)) Pipe to stdin"));
+      xdup2 (child_pipe[PREAD], STDIN_FILENO);
       xclose (child_pipe[PWRITE]);
       execlp (use_compress_program_option, use_compress_program_option,
 	      "-d", (char *) 0);
-      FATAL_ERROR ((0, errno, _("Cannot exec %s"),
-		    use_compress_program_option));
+      exec_fatal (use_compress_program_option);
     }
 
   /* The new born grandchild tar is here!  */
@@ -581,7 +567,7 @@ child_open_for_uncompress (void)
 
   /* Prepare for unblocking the data from the archive into the uncompressor.  */
 
-  xdup2 (child_pipe[PWRITE], STDOUT_FILENO, _("(grandchild) Pipe to stdout"));
+  xdup2 (child_pipe[PWRITE], STDOUT_FILENO);
   xclose (child_pipe[PREAD]);
 
   if (strcmp (archive_name_array[0], "-") == 0)
@@ -590,8 +576,7 @@ child_open_for_uncompress (void)
     archive = rmtopen (archive_name_array[0], O_RDONLY | O_BINARY,
 		       MODE_RW, rsh_command_option);
   if (archive < 0)
-    FATAL_ERROR ((0, errno, _("Cannot open archive %s"),
-		  archive_name_array[0]));
+    open_fatal (archive_name_array[0]);
 
   /* Let's read the archive and pipe it into stdout.  */
 
@@ -608,7 +593,7 @@ child_open_for_uncompress (void)
       status = rmtread (archive, record_start->buffer, record_size);
       if (status < 0)
 	{
-	  read_error ();
+	  archive_read_error ();
 	  goto error_loop;
 	}
       if (status == 0)
@@ -620,13 +605,11 @@ child_open_for_uncompress (void)
 	  count = maximum < BLOCKSIZE ? maximum : BLOCKSIZE;
 	  status = full_write (STDOUT_FILENO, cursor, count);
 	  if (status < 0)
-	    FATAL_ERROR ((0, errno, _("\
-Cannot write to compression program")));
+	    write_error (use_compress_program_option);
 
 	  if (status != count)
 	    {
-	      ERROR ((0, 0, _("\
-Write to compression program short %lu bytes"),
+	      ERROR ((0, 0, _("Write to compression program short %lu bytes"),
 		      (unsigned long) (count - status)));
 	      count = status;
 	    }
@@ -677,7 +660,7 @@ check_label_pattern (union block *label)
 `------------------------------------------------------------------------*/
 
 void
-open_archive (enum access_mode access)
+open_archive (enum access_mode wanted_access)
 {
   int backed_up_flag = 0;
 
@@ -689,8 +672,8 @@ open_archive (enum access_mode access)
   if (archive_names == 0)
     FATAL_ERROR ((0, 0, _("No archive name given")));
 
-  current_file_name = NULL;
-  current_link_name = NULL;
+  current_file_name = 0;
+  current_link_name = 0;
 
   /* FIXME: According to POSIX.1, PATH_MAX may well not be a compile-time
      constant, and the value from sysconf (_SC_PATH_MAX) may well not be any
@@ -699,31 +682,29 @@ open_archive (enum access_mode access)
      allocation.  (Roland McGrath)  */
 
   if (!real_s_name)
-    real_s_name = (char *) xmalloc (PATH_MAX);
+    real_s_name = xmalloc (PATH_MAX);
   /* FIXME: real_s_name is never freed.  */
 
-  save_name = NULL;
+  save_name = 0;
 
   if (multi_volume_option)
     {
-      record_start
-	= (union block *) valloc (record_size + (2 * BLOCKSIZE));
+      if (verify_option)
+	FATAL_ERROR ((0, 0, _("Cannot verify multi-volume archives")));
+      record_start = valloc (record_size + (2 * BLOCKSIZE));
       if (record_start)
 	record_start += 2;
     }
   else
-    record_start = (union block *) valloc (record_size);
+    record_start = valloc (record_size);
   if (!record_start)
-    FATAL_ERROR ((0, 0, _("Could not allocate memory for blocking factor %d"),
+    FATAL_ERROR ((0, 0, _("Cannot allocate memory for blocking factor %d"),
 		  blocking_factor));
 
   current_block = record_start;
   record_end = record_start + blocking_factor;
   /* When updating the archive, we start with reading.  */
-  access_mode = access == ACCESS_UPDATE ? ACCESS_READ : access;
-
-  if (multi_volume_option && verify_option)
-    FATAL_ERROR ((0, 0, _("Cannot verify multi-volume archives")));
+  access_mode = wanted_access == ACCESS_UPDATE ? ACCESS_READ : wanted_access;
 
   if (use_compress_program_option)
     {
@@ -732,7 +713,7 @@ open_archive (enum access_mode access)
       if (verify_option)
 	FATAL_ERROR ((0, 0, _("Cannot verify compressed archives")));
 
-      switch (access)
+      switch (wanted_access)
 	{
 	case ACCESS_READ:
 	  child_open_for_uncompress ();
@@ -747,7 +728,8 @@ open_archive (enum access_mode access)
 	  break;
 	}
 
-      if (access == ACCESS_WRITE && strcmp (archive_name_array[0], "-") == 0)
+      if (wanted_access == ACCESS_WRITE
+	  && strcmp (archive_name_array[0], "-") == 0)
 	stdlis = stderr;
     }
   else if (strcmp (archive_name_array[0], "-") == 0)
@@ -756,7 +738,7 @@ open_archive (enum access_mode access)
       if (verify_option)
 	FATAL_ERROR ((0, 0, _("Cannot verify stdin/stdout archive")));
 
-      switch (access)
+      switch (wanted_access)
 	{
 	case ACCESS_READ:
 	  archive = STDIN_FILENO;
@@ -778,7 +760,7 @@ open_archive (enum access_mode access)
     archive = rmtopen (archive_name_array[0], O_RDWR | O_CREAT | O_BINARY,
 		       MODE_RW, rsh_command_option);
   else
-    switch (access)
+    switch (wanted_access)
       {
       case ACCESS_READ:
 	archive = rmtopen (archive_name_array[0], O_RDONLY | O_BINARY,
@@ -808,8 +790,8 @@ open_archive (enum access_mode access)
 
       if (backed_up_flag)
 	undo_last_backup ();
-      FATAL_ERROR ((0, saved_errno, _("Cannot open %s"),
-		    archive_name_array[0]));
+      errno = saved_errno;
+      open_fatal (archive_name_array[0]);
     }
 
 #if !MSDOS
@@ -822,8 +804,9 @@ open_archive (enum access_mode access)
     dev_null_output =
       (strcmp (archive_name_array[0], dev_null) == 0
        || (! _isrmt (archive)
-	   && stat (dev_null, &dev_null_stat) == 0
 	   && S_ISCHR (archive_stat.st_mode)
+	   && stat (dev_null, &dev_null_stat) == 0
+	   && S_ISCHR (dev_null_stat.st_mode)
 	   && archive_stat.st_rdev == dev_null_stat.st_rdev));
   }
 
@@ -841,7 +824,7 @@ open_archive (enum access_mode access)
   setmode (archive, O_BINARY);
 #endif
 
-  switch (access)
+  switch (wanted_access)
     {
     case ACCESS_READ:
     case ACCESS_UPDATE:
@@ -853,18 +836,19 @@ open_archive (enum access_mode access)
 	  union block *label = find_next_block ();
 
 	  if (!label)
-	    FATAL_ERROR ((0, 0, _("Archive not labelled to match `%s'"),
-			  volume_label_option));
+	    FATAL_ERROR ((0, 0, _("Archive not labeled to match %s"),
+			  quote (volume_label_option)));
 	  if (!check_label_pattern (label))
-	    FATAL_ERROR ((0, 0, _("Volume `%s' does not match `%s'"),
-			  label->header.name, volume_label_option));
+	    FATAL_ERROR ((0, 0, _("Volume %s does not match %s"),
+			  quote_n (0, label->header.name),
+			  quote_n (1, volume_label_option)));
 	}
       break;
 
     case ACCESS_WRITE:
       if (volume_label_option)
 	{
-	  memset ((void *) record_start, 0, BLOCKSIZE);
+	  memset (record_start, 0, BLOCKSIZE);
 	  if (multi_volume_option)
 	    sprintf (record_start->header.name, "%s Volume 1",
 		     volume_label_option);
@@ -874,7 +858,7 @@ open_archive (enum access_mode access)
 	  assign_string (&current_file_name, record_start->header.name);
 
 	  record_start->header.typeflag = GNUTYPE_VOLHDR;
-	  TIME_TO_OCT (time (0), record_start->header.mtime);
+	  TIME_TO_CHARS (start_time, record_start->header.mtime);
 	  finish_header (record_start);
 #if 0
 	  current_block++;
@@ -897,10 +881,9 @@ flush_write (void)
   if (checkpoint_option && !(++checkpoint % 10))
     WARN ((0, 0, _("Write checkpoint %d"), checkpoint));
 
-  if (!zerop_tarlong (tape_length_option)
-      && !lessp_tarlong (bytes_written, tape_length_option))
+  if (tape_length_option && tape_length_option <= bytes_written)
     {
-      errno = ENOSPC;		/* FIXME: errno should be read-only */
+      errno = ENOSPC;
       status = 0;
     }
   else if (dev_null_output)
@@ -908,12 +891,10 @@ flush_write (void)
   else
     status = write_archive_buffer ();
   if (status != record_size && !multi_volume_option)
-    write_error (status);
-  else if (totals_option)
-    add_to_tarlong (total_written, record_size);
+    archive_write_error (status);
 
   if (status > 0)
-    add_to_tarlong (bytes_written, status);
+    bytes_written += status;
 
   if (status == record_size)
     {
@@ -929,11 +910,7 @@ flush_write (void)
 	      return;
 	    }
 
-	  cursor = save_name;
-#if MSDOS
-	  if (cursor[1] == ':')
-	    cursor += 2;
-#endif
+	  cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
 	  while (*cursor == '/')
 	    cursor++;
 
@@ -948,14 +925,16 @@ flush_write (void)
 
   /* ENXIO is for the UNIX PC.  */
   if (status < 0 && errno != ENOSPC && errno != EIO && errno != ENXIO)
-    write_error (status);
+    archive_write_error (status);
 
   /* If error indicates a short write, we just move to the next tape.  */
 
   if (!new_volume (ACCESS_WRITE))
     return;
 
-  clear_tarlong (bytes_written);
+  if (totals_option)
+    prev_written += bytes_written;
+  bytes_written = 0;
 
   if (volume_label_option && real_s_name[0])
     {
@@ -972,9 +951,10 @@ flush_write (void)
 
   if (volume_label_option)
     {
-      memset ((void *) record_start, 0, BLOCKSIZE);
-      sprintf (record_start->header.name, "%s Volume %d", volume_label_option, volno);
-      TIME_TO_OCT (time (0), record_start->header.mtime);
+      memset (record_start, 0, BLOCKSIZE);
+      sprintf (record_start->header.name, "%s Volume %d",
+	       volume_label_option, volno);
+      TIME_TO_CHARS (start_time, record_start->header.mtime);
       record_start->header.typeflag = GNUTYPE_VOLHDR;
       finish_header (record_start);
     }
@@ -986,7 +966,7 @@ flush_write (void)
       if (volume_label_option)
 	record_start++;
 
-      memset ((void *) record_start, 0, BLOCKSIZE);
+      memset (record_start, 0, BLOCKSIZE);
 
       /* FIXME: Michael P Urban writes: [a long name file] is being written
 	 when a new volume rolls around [...]  Looks like the wrong value is
@@ -994,9 +974,9 @@ flush_write (void)
 
       strcpy (record_start->header.name, real_s_name);
       record_start->header.typeflag = GNUTYPE_MULTIVOL;
-      OFF_TO_OCT (real_s_sizeleft, record_start->header.size);
-      OFF_TO_OCT (real_s_totsize - real_s_sizeleft, 
-		  record_start->oldgnu_header.offset);
+      OFF_TO_CHARS (real_s_sizeleft, record_start->header.size);
+      OFF_TO_CHARS (real_s_totsize - real_s_sizeleft, 
+		    record_start->oldgnu_header.offset);
       tmp = verbose_option;
       verbose_option = 0;
       finish_header (record_start);
@@ -1008,17 +988,16 @@ flush_write (void)
 
   status = write_archive_buffer ();
   if (status != record_size)
-    write_error (status);
-  else if (totals_option)
-    add_to_tarlong (total_written, record_size);
+    archive_write_error (status);
+
+  bytes_written += status;
 
-  add_to_tarlong (bytes_written, record_size);
   if (copy_back)
     {
       record_start += copy_back;
-      memcpy ((void *) current_block,
-	      (void *) (record_start + blocking_factor - copy_back),
-	      (size_t) (copy_back * BLOCKSIZE));
+      memcpy (current_block,
+	      record_start + blocking_factor - copy_back,
+	      copy_back * BLOCKSIZE);
       current_block += copy_back;
 
       if (real_s_sizeleft >= copy_back * BLOCKSIZE)
@@ -1027,12 +1006,8 @@ flush_write (void)
 	real_s_name[0] = '\0';
       else
 	{
-	  char *cursor = save_name;
+	  char *cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
 
-#if MSDOS
-	  if (cursor[1] == ':')
-	    cursor += 2;
-#endif
 	  while (*cursor == '/')
 	    cursor++;
 
@@ -1051,22 +1026,18 @@ flush_write (void)
 `---------------------------------------------------------------------*/
 
 static void
-write_error (ssize_t status)
+archive_write_error (ssize_t status)
 {
-  int saved_errno = errno;
-
   /* It might be useful to know how much was written before the error
-     occured.  Beware that mere printing maybe change errno value.  */
+     occurred.  */
   if (totals_option)
-    print_total_written ();
+    {
+      int e = errno;
+      print_total_written ();
+      errno = e;
+    }
 
-  if (status < 0)
-    FATAL_ERROR ((0, saved_errno, _("Cannot write to %s"),
-		  *archive_name_cursor));
-  else
-    FATAL_ERROR ((0, 0, _("Only wrote %lu of %lu bytes to %s"),
-		  (unsigned long) status, (unsigned long) record_size,
-		  *archive_name_cursor));
+  write_fatal_details (*archive_name_cursor, status, record_size);
 }
 
 /*-------------------------------------------------------------------.
@@ -1075,9 +1046,9 @@ write_error (ssize_t status)
 `-------------------------------------------------------------------*/
 
 static void
-read_error (void)
+archive_read_error (void)
 {
-  WARN ((0, errno, _("Read error on %s"), *archive_name_cursor));
+  read_error (*archive_name_cursor);
 
   if (record_start_block == 0)
     FATAL_ERROR ((0, 0, _("At beginning of tape, quitting now")));
@@ -1111,20 +1082,18 @@ flush_read (void)
 
   if (write_archive_to_stdout && record_start_block != 0)
     {
+      archive = STDOUT_FILENO;
       status = write_archive_buffer ();
+      archive = STDIN_FILENO;
       if (status != record_size)
-	write_error (status);
+	archive_write_error (status);
     }
   if (multi_volume_option)
     {
       if (save_name)
 	{
-	  char *cursor = save_name;
+	  char *cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
 
-#if MSDOS
-	  if (cursor[1] == ':')
-	    cursor += 2;
-#endif
 	  while (*cursor == '/')
 	    cursor++;
 
@@ -1140,7 +1109,7 @@ flush_read (void)
 	}
     }
 
-error_loop:
+ error_loop:
   status = rmtread (archive, record_start->buffer, record_size);
   if (status == record_size)
     return;
@@ -1172,7 +1141,7 @@ error_loop:
       status = rmtread (archive, record_start->buffer, record_size);
       if (status < 0)
 	{
-	  read_error ();
+	  archive_read_error ();
 	  goto vol_error;
 	}
       if (status != record_size)
@@ -1186,15 +1155,16 @@ error_loop:
 	    {
 	      if (!check_label_pattern (cursor))
 		{
-		  WARN ((0, 0, _("Volume `%s' does not match `%s'"),
-			 cursor->header.name, volume_label_option));
+		  WARN ((0, 0, _("Volume %s does not match %s"),
+			 quote_n (0, cursor->header.name),
+			 quote_n (1, volume_label_option)));
 		  volno--;
 		  global_volno--;
 		  goto try_volume;
 		}
 	    }
 	  if (verbose_option)
-	    fprintf (stdlis, _("Reading %s\n"), cursor->header.name);
+	    fprintf (stdlis, _("Reading %s\n"), quote (cursor->header.name));
 	  cursor++;
 	}
       else if (volume_label_option)
@@ -1207,13 +1177,13 @@ error_loop:
 	      || strcmp (cursor->header.name, real_s_name))
 	    {
 	      WARN ((0, 0, _("%s is not continued on this volume"),
-		     real_s_name));
+		     quote (real_s_name)));
 	      volno--;
 	      global_volno--;
 	      goto try_volume;
 	    }
-	  s1 = UINTMAX_FROM_OCT (cursor->header.size);
-	  s2 = UINTMAX_FROM_OCT (cursor->oldgnu_header.offset);
+	  s1 = UINTMAX_FROM_HEADER (cursor->header.size);
+	  s2 = UINTMAX_FROM_HEADER (cursor->oldgnu_header.offset);
 	  if (real_s_totsize != s1 + s2 || s1 + s2 < s2)
 	    {
 	      char totsizebuf[UINTMAX_STRSIZE_BOUND];
@@ -1221,7 +1191,7 @@ error_loop:
 	      char s2buf[UINTMAX_STRSIZE_BOUND];
 	      
 	      WARN ((0, 0, _("%s is the wrong size (%s != %s + %s)"),
-		     cursor->header.name,
+		     quote (cursor->header.name),
 		     STRINGIFY_BIGINT (save_totsize, totsizebuf),
 		     STRINGIFY_BIGINT (s1, s1buf),
 		     STRINGIFY_BIGINT (s2, s2buf)));
@@ -1230,7 +1200,7 @@ error_loop:
 	      goto try_volume;
 	    }
 	  if (real_s_totsize - real_s_sizeleft
-	      != OFF_FROM_OCT (cursor->oldgnu_header.offset))
+	      != OFF_FROM_HEADER (cursor->oldgnu_header.offset))
 	    {
 	      WARN ((0, 0, _("This volume is out of sequence")));
 	      volno--;
@@ -1244,53 +1214,46 @@ error_loop:
     }
   else if (status < 0)
     {
-      read_error ();
+      archive_read_error ();
       goto error_loop;		/* try again */
     }
 
-short_read:
+ short_read:
   more = record_start->buffer + status;
   left = record_size - status;
 
-again:
-  if (left % BLOCKSIZE == 0)
+  while (left % BLOCKSIZE != 0)
     {
-      /* FIXME: for size=0, multi-volume support.  On the first record, warn
-	 about the problem.  */
+      if (status)
+	while ((status = rmtread (archive, more, left)) < 0)
+	  archive_read_error ();
 
-      if (!read_full_records_option && verbose_option
-	  && record_start_block == 0 && status > 0)
-	WARN ((0, 0, _("Record size = %lu blocks"),
-	       (unsigned long) (status / BLOCKSIZE)));
+      if (status == 0)
+	{
+	  ERROR ((0, 0, _("%d garbage bytes ignored at end of archive"),
+		  (int) ((record_size - left) % BLOCKSIZE)));
+	  break;
+	}
 
-      record_end = record_start + (record_size - left) / BLOCKSIZE;
-
-      return;
-    }
-  if (read_full_records_option)
-    {
+      if (! read_full_records_option)
+	FATAL_ERROR ((0, 0, _("Unaligned block (%lu bytes) in archive"),
+		      (unsigned long) (record_size - left)));
+	  
       /* User warned us about this.  Fix up.  */
 
-      if (left > 0)
-	{
-	error2loop:
-	  status = rmtread (archive, more, left);
-	  if (status < 0)
-	    {
-	      read_error ();
-	      goto error2loop;	/* try again */
-	    }
-	  if (status == 0)
-	    FATAL_ERROR ((0, 0, _("Archive %s EOF not on block boundary"),
-			  *archive_name_cursor));
-	  left -= status;
-	  more += status;
-	  goto again;
-	}
+      left -= status;
+      more += status;
     }
-  else
-    FATAL_ERROR ((0, 0, _("Only read %lu bytes from archive %s"),
-		  (unsigned long) status, *archive_name_cursor));
+
+  /* FIXME: for size=0, multi-volume support.  On the first record, warn
+     about the problem.  */
+
+  if (!read_full_records_option && verbose_option
+      && record_start_block == 0 && status > 0)
+    WARN ((0, 0, _("Record size = %lu blocks"),
+	   (unsigned long) ((record_size - left) / BLOCKSIZE)));
+
+  record_end = record_start + (record_size - left) / BLOCKSIZE;
 }
 
 /*-----------------------------------------------.
@@ -1311,11 +1274,8 @@ flush_archive (void)
 
       if (file_to_switch_to >= 0)
 	{
-	  int status = rmtclose (archive);
-
-	  if (status < 0)
-	    WARN ((0, errno, _("WARNING: Cannot close %s (%d, %d)"),
-		   *archive_name_cursor, archive, status));
+	  if (rmtclose (archive) != 0)
+	    close_warn (*archive_name_cursor);
 
 	  archive = file_to_switch_to;
 	}
@@ -1370,14 +1330,14 @@ backspace_output (void)
       {
 	/* Lseek failed.  Try a different method.  */
 
-	WARN ((0, 0, _("\
-Could not backspace archive file; it may be unreadable without -i")));
+	WARN ((0, 0,
+	       _("Cannot backspace archive file; it may be unreadable without -i")));
 
 	/* Replace the first part of the record with NULs.  */
 
 	if (record_start->buffer != output_start)
 	  memset (record_start->buffer, 0,
-		  (size_t) (output_start - record_start->buffer));
+		  output_start - record_start->buffer);
       }
   }
 }
@@ -1415,58 +1375,44 @@ close_archive (void)
       int status = pos < 0 ? -1 : ftruncate (archive, pos);
 #endif
       if (status != 0)
-	WARN ((0, errno, _("WARNING: Cannot truncate %s"),
-	       *archive_name_cursor));
+	truncate_warn (*archive_name_cursor);
     }
   if (verify_option)
     verify_volume ();
 
-  {
-    int status = rmtclose (archive);
-
-    if (status < 0)
-      WARN ((0, errno, _("WARNING: Cannot close %s (%d, %d)"),
-	     *archive_name_cursor, archive, status));
-  }
+  if (rmtclose (archive) != 0)
+    close_warn (*archive_name_cursor);
 
 #if !MSDOS
 
   if (child_pid)
     {
-      WAIT_T wait_status;
-      pid_t child;
+      int wait_status;
 
-      /* Loop waiting for the right child to die, or for no more kids.  */
+      while (waitpid (child_pid, &wait_status, 0) == -1)
+	if (errno != EINTR)
+	  {
+	    waitpid_error (use_compress_program_option);
+	    break;
+	  }
 
-      while ((child = wait (&wait_status), child != child_pid)
-	     && child != -1)
-	continue;
-
-      if (child != -1)
+      if (WIFSIGNALED (wait_status))
 	{
-	  if (WIFSIGNALED (wait_status)
-#if 0
-	      && !WIFSTOPPED (wait_status)
-#endif
-	      )
-	    {
-	      /* SIGPIPE is OK, everything else is a problem.  */
-
-	      if (WTERMSIG (wait_status) != SIGPIPE)
-		ERROR ((0, 0, _("Child died with signal %d%s"),
-			WTERMSIG (wait_status),
-			WCOREDUMP (wait_status) ? _(" (core dumped)") : ""));
-	    }
-	  else
-	    {
-	      /* Child voluntarily terminated -- but why?  /bin/sh returns
-		 SIGPIPE + 128 if its child, then do nothing.  */
-
-	      if (WEXITSTATUS (wait_status) != (SIGPIPE + 128)
-		  && WEXITSTATUS (wait_status))
-		ERROR ((0, 0, _("Child returned status %d"),
-			WEXITSTATUS (wait_status)));
-	    }
+	  /* SIGPIPE is OK, everything else is a problem.  */
+	  
+	  if (WTERMSIG (wait_status) != SIGPIPE)
+	    ERROR ((0, 0, _("Child died with signal %d"),
+		    WTERMSIG (wait_status)));
+	}
+      else
+	{
+	  /* Child voluntarily terminated -- but why?  /bin/sh returns
+	     SIGPIPE + 128 if its child, then do nothing.  */
+	      
+	  if (WEXITSTATUS (wait_status)
+	      && WEXITSTATUS (wait_status) != (SIGPIPE + 128))
+	    ERROR ((0, 0, _("Child returned status %d"),
+		    WEXITSTATUS (wait_status)));
 	}
     }
 #endif /* !MSDOS */
@@ -1492,11 +1438,13 @@ init_volume_number (void)
   if (file)
     {
       fscanf (file, "%d", &global_volno);
-      if (fclose (file) == EOF)
-	ERROR ((0, errno, "%s", volno_file_option));
+      if (ferror (file))
+	read_error (volno_file_option);
+      if (fclose (file) != 0)
+	close_error (volno_file_option);
     }
   else if (errno != ENOENT)
-    ERROR ((0, errno, "%s", volno_file_option));
+    open_error (volno_file_option);
 }
 
 /*-------------------------------------------------------.
@@ -1511,11 +1459,13 @@ closeout_volume_number (void)
   if (file)
     {
       fprintf (file, "%d\n", global_volno);
-      if (fclose (file) == EOF)
-	ERROR ((0, errno, "%s", volno_file_option));
+      if (ferror (file))
+	write_error (volno_file_option);
+      if (fclose (file) != 0)
+	close_error (volno_file_option);
     }
   else
-    ERROR ((0, errno, "%s", volno_file_option));
+    open_error (volno_file_option);
 }
 
 /*-----------------------------------------------------------------------.
@@ -1526,10 +1476,8 @@ closeout_volume_number (void)
 static int
 new_volume (enum access_mode access)
 {
-  static FILE *read_file = NULL;
-  static int looped = 0;
-
-  int status;
+  static FILE *read_file;
+  static int looped;
 
   if (!read_file && !info_script_option)
     /* FIXME: if fopen is used, it will never be closed.  */
@@ -1540,9 +1488,8 @@ new_volume (enum access_mode access)
   if (verify_option)
     verify_volume ();
 
-  if (status = rmtclose (archive), status < 0)
-    WARN ((0, errno, _("WARNING: Cannot close %s (%d, %d)"),
-	   *archive_name_cursor, archive, status));
+  if (rmtclose (archive) != 0)
+    close_warn (*archive_name_cursor);
 
   global_volno++;
   volno++;
@@ -1553,7 +1500,7 @@ new_volume (enum access_mode access)
       looped = 1;
     }
 
-tryagain:
+ tryagain:
   if (looped)
     {
       /* We have to prompt from now on.  */
@@ -1572,19 +1519,19 @@ tryagain:
 	    fputc ('\007', stderr);
 	    fprintf (stderr,
 		     _("Prepare volume #%d for %s and hit return: "),
-		     global_volno, *archive_name_cursor);
+		     global_volno, quote (*archive_name_cursor));
 	    fflush (stderr);
 
-	    if (fgets (input_buffer, sizeof (input_buffer), read_file) == 0)
+	    if (fgets (input_buffer, sizeof input_buffer, read_file) == 0)
 	      {
-		fprintf (stderr, _("EOF where user reply was expected"));
+		WARN ((0, 0, _("EOF where user reply was expected")));
 
 		if (subcommand_option != EXTRACT_SUBCOMMAND
 		    && subcommand_option != LIST_SUBCOMMAND
 		    && subcommand_option != DIFF_SUBCOMMAND)
 		  WARN ((0, 0, _("WARNING: Archive is incomplete")));
 
-		exit (TAREXIT_FAILURE);
+		fatal_exit ();
 	      }
 	    if (input_buffer[0] == '\n'
 		|| input_buffer[0] == 'y'
@@ -1606,14 +1553,14 @@ tryagain:
 	      case 'q':
 		/* Quit.  */
 
-		fprintf (stdlis, _("No new volume; exiting.\n"));
+		WARN ((0, 0, _("No new volume; exiting.\n")));
 
 		if (subcommand_option != EXTRACT_SUBCOMMAND
 		    && subcommand_option != LIST_SUBCOMMAND
 		    && subcommand_option != DIFF_SUBCOMMAND)
 		  WARN ((0, 0, _("WARNING: Archive is incomplete")));
 
-		exit (TAREXIT_FAILURE);
+		fatal_exit ();
 
 	      case 'n':
 		/* Get new file name.  */
@@ -1638,35 +1585,28 @@ tryagain:
 #if MSDOS
 		spawnl (P_WAIT, getenv ("COMSPEC"), "-", 0);
 #else /* not MSDOS */
-		switch (fork ())
-		  {
-		  case -1:
-		    WARN ((0, errno, _("Cannot fork!")));
-		    break;
-
-		  case 0:
+		{
+		  pid_t child;
+		  const char *shell = getenv ("SHELL");
+		  if (! shell)
+		    shell = "/bin/sh";
+		  child = xfork ();
+		  if (child == 0)
 		    {
-		      const char *shell = getenv ("SHELL");
-
-		      if (shell == NULL)
-			shell = "/bin/sh";
 		      execlp (shell, "-sh", "-i", 0);
-		      FATAL_ERROR ((0, errno, _("Cannot exec a shell %s"),
-				    shell));
+		      exec_fatal (shell);
 		    }
-
-		  default:
+		  else
 		    {
-		      WAIT_T wait_status;
-
-		      wait (&wait_status);
+		      int wait_status;
+		      while (waitpid (child, &wait_status, 0) == -1)
+			if (errno != EINTR)
+			  {
+			    waitpid_error (shell);
+			    break;
+			  }
 		    }
-		    break;
-		  }
-
-		/* FIXME: I'm not sure if that's all that has to be done
-		   here.  (jk)  */
-
+		}
 #endif /* not MSDOS */
 		break;
 	      }
@@ -1699,7 +1639,7 @@ tryagain:
 
   if (archive < 0)
     {
-      WARN ((0, errno, _("Cannot open %s"), *archive_name_cursor));
+      open_warn (*archive_name_cursor);
       if (!verify_option && access == ACCESS_WRITE && backup_option)
 	undo_last_backup ();
       goto tryagain;
