@@ -1,5 +1,5 @@
 /* Diff files from a tar archive.
-   Copyright (C) 1988, 92, 93, 94, 96, 97, 1999 Free Software Foundation, Inc.
+   Copyright 1988, 92, 93, 94, 96, 97, 99, 2000 Free Software Foundation, Inc.
    Written by John Gilmore, on 1987-04-30.
 
    This program is free software; you can redistribute it and/or modify it
@@ -18,10 +18,22 @@
 
 #include "system.h"
 
+#if HAVE_UTIME_H
+# include <utime.h>
+#else
+struct utimbuf
+  {
+    long actime;
+    long modtime;
+  };
+#endif
+
 #if HAVE_LINUX_FD_H
 # include <linux/fd.h>
 #endif
 
+#include <quotearg.h>
+
 #include "common.h"
 #include "rmt.h"
 
@@ -29,13 +41,13 @@
 #define MESSAGE_BUFFER_SIZE 100
 
 /* Nonzero if we are verifying at the moment.  */
-int now_verifying = 0;
+int now_verifying;
 
 /* File descriptor for the file we are diffing.  */
 static int diff_handle;
 
 /* Area for reading file contents into.  */
-static char *diff_buffer = NULL;
+static char *diff_buffer;
 
 /*--------------------------------.
 | Initialize for a diff operation |
@@ -44,23 +56,21 @@ static char *diff_buffer = NULL;
 void
 diff_init (void)
 {
-  diff_buffer = (char *) valloc (record_size);
+  diff_buffer = valloc (record_size);
   if (!diff_buffer)
-    FATAL_ERROR ((0, 0,
-		  _("Could not allocate memory for diff buffer of %lu bytes"),
-		  (unsigned long) record_size));
+    xalloc_die ();
 }
 
 /*------------------------------------------------------------------------.
 | Sigh about something that differs by writing a MESSAGE to stdlis, given |
-| MESSAGE is not NULL.  Also set the exit status if not already.          |
+| MESSAGE is nonzero.  Also set the exit status if not already.           |
 `------------------------------------------------------------------------*/
 
 static void
 report_difference (const char *message)
 {
   if (message)
-    fprintf (stdlis, "%s: %s\n", current_file_name, message);
+    fprintf (stdlis, "%s: %s\n", quotearg_colon (current_file_name), message);
 
   if (exit_status == TAREXIT_SUCCESS)
     exit_status = TAREXIT_DIFFERS;
@@ -93,8 +103,8 @@ process_rawdata (size_t bytes, char *buffer)
     {
       if (status < 0)
 	{
-	  WARN ((0, errno, _("Cannot read %s"), current_file_name));
-	  report_difference (NULL);
+	  read_error (current_file_name);
+	  report_difference (0);
 	}
       else
 	{
@@ -107,7 +117,7 @@ process_rawdata (size_t bytes, char *buffer)
 
   if (memcmp (buffer, diff_buffer, bytes))
     {
-      report_difference (_("Data differs"));
+      report_difference (_("Contents differ"));
       return 0;
     }
 
@@ -127,7 +137,7 @@ process_dumpdir (size_t bytes, char *buffer)
 {
   if (memcmp (buffer, dumpdir_cursor, bytes))
     {
-      report_difference (_("Data differs"));
+      report_difference (_("Contents differ"));
       return 0;
     }
 
@@ -154,9 +164,9 @@ read_and_process (size_t size, int (*processor) (size_t, char *))
   while (size)
     {
       data_block = find_next_block ();
-      if (data_block == NULL)
+      if (! data_block)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF on archive file")));
+	  ERROR ((0, 0, _("Unexpected EOF in archive")));
 	  return;
 	}
 
@@ -181,7 +191,7 @@ read_and_process (size_t size, int (*processor) (size_t, char *))
    that.  Anyhow, what it does is translate the sparse information on the
    header, and in any subsequent extended headers, into an array of
    structures with true numbers, as opposed to character strings.  It
-   simply makes our life much easier, doing so many comparisong and such.
+   simply makes our life much easier, doing so many comparisons and such.
    */
 
 static void
@@ -193,7 +203,7 @@ fill_in_sparse_array (void)
      long, but can change in this routine if necessary.  */
 
   sp_array_size = 10;
-  sparsearray = (struct sp_array *) xmalloc (sp_array_size * sizeof (struct sp_array));
+  sparsearray = xmalloc (sp_array_size * sizeof (struct sp_array));
 
   /* There are at most five of these structures in the header itself;
      read these in first.  */
@@ -205,9 +215,9 @@ fill_in_sparse_array (void)
 	break;
 
       sparsearray[counter].offset =
-	OFF_FROM_OCT (current_header->oldgnu_header.sp[counter].offset);
+	OFF_FROM_HEADER (current_header->oldgnu_header.sp[counter].offset);
       sparsearray[counter].numbytes =
-	SIZE_FROM_OCT (current_header->oldgnu_header.sp[counter].numbytes);
+	SIZE_FROM_HEADER (current_header->oldgnu_header.sp[counter].numbytes);
     }
 
   /* If the header's extended, we gotta read in exhdr's till we're done.  */
@@ -221,6 +231,9 @@ fill_in_sparse_array (void)
       while (1)
 	{
 	  exhdr = find_next_block ();
+	  if (!exhdr)
+	    FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
+
 	  for (counter = 0; counter < SPARSES_IN_SPARSE_HEADER; counter++)
 	    {
 	      if (counter + so_far_ind > sp_array_size - 1)
@@ -229,7 +242,7 @@ fill_in_sparse_array (void)
 		     realloc it.  */
 
 		  sp_array_size *= 2;
-		  sparsearray = (struct sp_array *)
+		  sparsearray =
 		    xrealloc (sparsearray,
 			      sp_array_size * sizeof (struct sp_array));
 		}
@@ -237,9 +250,9 @@ fill_in_sparse_array (void)
 	      /* Convert the character strings into offsets and sizes.  */
 
 	      sparsearray[counter + so_far_ind].offset =
-		OFF_FROM_OCT (exhdr->sparse_header.sp[counter].offset);
+		OFF_FROM_HEADER (exhdr->sparse_header.sp[counter].offset);
 	      sparsearray[counter + so_far_ind].numbytes =
-		SIZE_FROM_OCT (exhdr->sparse_header.sp[counter].numbytes);
+		SIZE_FROM_HEADER (exhdr->sparse_header.sp[counter].numbytes);
 	    }
 
 	  /* If this is the last extended header for this file, we can
@@ -264,7 +277,7 @@ fill_in_sparse_array (void)
 
 /* JK Diff'ing a sparse file with its counterpart on the tar file is a
    bit of a different story than a normal file.  First, we must know what
-   areas of the file to skip through, i.e., we need to contruct a
+   areas of the file to skip through, i.e., we need to construct a
    sparsearray, which will hold all the information we need.  We must
    compare small amounts of data at a time as we find it.  */
 
@@ -276,9 +289,9 @@ static void
 diff_sparse_files (off_t size_of_file)
 {
   off_t remaining_size = size_of_file;
-  char *buffer = (char *) xmalloc (BLOCKSIZE * sizeof (char));
+  char *buffer = xmalloc (BLOCKSIZE * sizeof (char));
   size_t buffer_size = BLOCKSIZE;
-  union block *data_block = NULL;
+  union block *data_block = 0;
   int counter = 0;
   int different = 0;
 
@@ -295,6 +308,8 @@ diff_sparse_files (off_t size_of_file)
 #endif
 
       data_block = find_next_block ();
+      if (!data_block)
+	FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
       chunk_size = sparsearray[counter].numbytes;
       if (!chunk_size)
 	break;
@@ -302,10 +317,8 @@ diff_sparse_files (off_t size_of_file)
       offset = sparsearray[counter].offset;
       if (lseek (diff_handle, offset, SEEK_SET) < 0)
 	{
-	  char buf[UINTMAX_STRSIZE_BOUND];
-	  WARN ((0, errno, _("Cannot seek to %s in file %s"),
-		 STRINGIFY_BIGINT (offset, buf), current_file_name));
-	  report_difference (NULL);
+	  seek_error_details (current_file_name, offset);
+	  report_difference (0);
 	}
 
       /* Take care to not run out of room in our buffer.  */
@@ -313,9 +326,9 @@ diff_sparse_files (off_t size_of_file)
       while (buffer_size < chunk_size)
 	{
 	  if (buffer_size * 2 < buffer_size)
-	    FATAL_ERROR ((0, 0, _("Memory exhausted")));
+	    xalloc_die ();
 	  buffer_size *= 2;
-	  buffer = (char *) xrealloc (buffer, buffer_size * sizeof (char));
+	  buffer = xrealloc (buffer, buffer_size * sizeof (char));
 	}
 
       while (chunk_size > BLOCKSIZE)
@@ -325,8 +338,8 @@ diff_sparse_files (off_t size_of_file)
 	    {
 	      if (status < 0)
 		{
-		  WARN ((0, errno, _("Cannot read %s"), current_file_name));
-		  report_difference (NULL);
+		  read_error (current_file_name);
+		  report_difference (0);
 		}
 	      else
 		{
@@ -349,14 +362,16 @@ diff_sparse_files (off_t size_of_file)
 	  remaining_size -= status;
 	  set_next_block_after (data_block);
 	  data_block = find_next_block ();
+	  if (!data_block)
+	    FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
 	}
       if (status = safe_read (diff_handle, buffer, chunk_size),
 	  status != chunk_size)
 	{
 	  if (status < 0)
 	    {
-	      WARN ((0, errno, _("Cannot read %s"), current_file_name));
-	      report_difference (NULL);
+	      read_error (current_file_name);
+	      report_difference (0);
 	    }
 	  else
 	    {
@@ -381,6 +396,8 @@ diff_sparse_files (off_t size_of_file)
 	  amount_read = 0;
 	  set_next_block_after (data_block);
 	  data_block = find_next_block ();
+	  if (!data_block)
+	    FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
 	}
 #endif
       set_next_block_after (data_block);
@@ -400,7 +417,7 @@ diff_sparse_files (off_t size_of_file)
   free (sparsearray);
 
   if (different)
-    report_difference (_("Data differs"));
+    report_difference (_("Contents differ"));
 }
 
 /*---------------------------------------------------------------------.
@@ -410,24 +427,17 @@ diff_sparse_files (off_t size_of_file)
 `---------------------------------------------------------------------*/
 
 static int
-get_stat_data (struct stat *stat_data)
+get_stat_data (char const *file_name, struct stat *stat_data)
 {
-  int status = (dereference_option
-		? stat (current_file_name, stat_data)
-		: lstat (current_file_name, stat_data));
+  int status = deref_stat (dereference_option, file_name, stat_data);
 
-  if (status < 0)
+  if (status != 0)
     {
       if (errno == ENOENT)
-	report_difference (_("File does not exist"));
+	stat_warn (file_name);
       else
-	{
-	  ERROR ((0, errno, _("Cannot stat file %s"), current_file_name));
-	  report_difference (NULL);
-	}
-#if 0
-      skip_file (current_stat.st_size);
-#endif
+	stat_error (file_name);
+      report_difference (0);
       return 0;
     }
 
@@ -444,9 +454,7 @@ diff_archive (void)
   struct stat stat_data;
   size_t name_length;
   int status;
-
-  errno = EPIPE;		/* FIXME: errno should be read-only */
-				/* FIXME: remove perrors */
+  struct utimbuf restore_times;
 
   set_next_block_after (current_header);
   decode_header (current_header, &current_stat, &current_format, 1);
@@ -463,8 +471,9 @@ diff_archive (void)
   switch (current_header->header.typeflag)
     {
     default:
-      WARN ((0, 0, _("Unknown file type '%c' for %s, diffed as normal file"),
-		 current_header->header.typeflag, current_file_name));
+      ERROR ((0, 0, _("%s: Unknown file type '%c', diffed as normal file"),
+	      quotearg_colon (current_file_name),
+	      current_header->header.typeflag));
       /* Fall through.  */
 
     case AREGTYPE:
@@ -478,7 +487,7 @@ diff_archive (void)
       if (current_file_name[name_length] == '/')
 	goto really_dir;
 
-      if (!get_stat_data (&stat_data))
+      if (!get_stat_data (current_file_name, &stat_data))
 	{
 	  if (current_header->oldgnu_header.isextended)
 	    skip_extended_headers ();
@@ -488,13 +497,12 @@ diff_archive (void)
 
       if (!S_ISREG (stat_data.st_mode))
 	{
-	  report_difference (_("Not a regular file"));
+	  report_difference (_("File type differs"));
 	  skip_file (current_stat.st_size);
 	  goto quit;
 	}
 
-      stat_data.st_mode &= MODE_ALL;
-      if (stat_data.st_mode != current_stat.st_mode)
+      if ((current_stat.st_mode & MODE_ALL) != (stat_data.st_mode & MODE_ALL))
 	report_difference (_("Mode differs"));
 
 #if !MSDOS
@@ -517,27 +525,21 @@ diff_archive (void)
 	  goto quit;
 	}
 
-      diff_handle = open (current_file_name, O_NDELAY | O_RDONLY | O_BINARY);
+      diff_handle = open (current_file_name, O_RDONLY | O_BINARY);
 
-      if (diff_handle < 0 && !absolute_names_option)
-	{
-	  char *tmpbuf = xmalloc (strlen (current_file_name) + 2);
-
-	  *tmpbuf = '/';
-	  strcpy (tmpbuf + 1, current_file_name);
-	  diff_handle = open (tmpbuf, O_NDELAY | O_RDONLY);
-	  free (tmpbuf);
-	}
       if (diff_handle < 0)
 	{
-	  ERROR ((0, errno, _("Cannot open %s"), current_file_name));
+	  open_error (current_file_name);
 	  if (current_header->oldgnu_header.isextended)
 	    skip_extended_headers ();
 	  skip_file (current_stat.st_size);
-	  report_difference (NULL);
+	  report_difference (0);
 	  goto quit;
 	}
 
+      restore_times.actime = stat_data.st_atime;
+      restore_times.modtime = stat_data.st_mtime;
+
       /* Need to treat sparse files completely differently here.  */
 
       if (current_header->header.typeflag == GNUTYPE_SPARSE)
@@ -554,12 +556,15 @@ diff_archive (void)
 	  read_and_process (current_stat.st_size, process_rawdata);
 
 	  if (multi_volume_option)
-	    assign_string (&save_name, NULL);
+	    assign_string (&save_name, 0);
 	}
 
       status = close (diff_handle);
-      if (status < 0)
-	ERROR ((0, errno, _("Error while closing %s"), current_file_name));
+      if (status != 0)
+	close_error (current_file_name);
+
+      if (atime_preserve_option)
+	utime (current_file_name, &restore_times);
 
     quit:
       break;
@@ -567,33 +572,21 @@ diff_archive (void)
 #if !MSDOS
     case LNKTYPE:
       {
-	dev_t dev;
-	ino_t ino;
+	struct stat link_data;
 
-	if (!get_stat_data (&stat_data))
+	if (!get_stat_data (current_file_name, &stat_data))
+	  break;
+	if (!get_stat_data (current_link_name, &link_data))
 	  break;
 
-	dev = stat_data.st_dev;
-	ino = stat_data.st_ino;
-	status = stat (current_link_name, &stat_data);
-	if (status < 0)
+	if (stat_data.st_dev != link_data.st_dev
+	    || stat_data.st_ino != link_data.st_ino)
 	  {
-	    if (errno == ENOENT)
-	      report_difference (_("Does not exist"));
-	    else
-	      {
-		WARN ((0, errno, _("Cannot stat file %s"), current_file_name));
-		report_difference (NULL);
-	      }
-	    break;
-	  }
+	    char *message =
+	      xmalloc (MESSAGE_BUFFER_SIZE + 4 * strlen (current_link_name));
 
-	if (stat_data.st_dev != dev || stat_data.st_ino != ino)
-	  {
-	    char *message = (char *)
-	      xmalloc (MESSAGE_BUFFER_SIZE + strlen (current_link_name));
-
-	    sprintf (message, _("Not linked to %s"), current_link_name);
+	    sprintf (message, _("Not linked to %s"),
+		     quote (current_link_name));
 	    report_difference (message);
 	    free (message);
 	    break;
@@ -603,7 +596,7 @@ diff_archive (void)
       }
 #endif /* not MSDOS */
 
-#ifdef S_ISLNK
+#ifdef HAVE_READLINK
     case SYMTYPE:
       {
 	char linkbuf[NAME_FIELD_SIZE + 3]; /* FIXME: may be too short.  */
@@ -613,71 +606,52 @@ diff_archive (void)
 	if (status < 0)
 	  {
 	    if (errno == ENOENT)
-	      report_difference (_("No such file or directory"));
+	      readlink_warn (current_file_name);
 	    else
-	      {
-		WARN ((0, errno, _("Cannot read link %s"), current_file_name));
-		report_difference (NULL);
-	      }
+	      readlink_error (current_file_name);
+	    report_difference (0);
 	    break;
 	  }
 
 	linkbuf[status] = '\0';	/* null-terminate it */
-	if (strncmp (current_link_name, linkbuf, (size_t) status) != 0)
+	if (strncmp (current_link_name, linkbuf, status) != 0)
 	  report_difference (_("Symlink differs"));
 
 	break;
       }
-#endif /* not S_ISLNK */
+#endif
 
-#ifdef S_IFCHR
     case CHRTYPE:
-      current_stat.st_mode |= S_IFCHR;
-      goto check_node;
-#endif /* not S_IFCHR */
-
-#ifdef S_IFBLK
-      /* If local system doesn't support block devices, use default case.  */
-
     case BLKTYPE:
-      current_stat.st_mode |= S_IFBLK;
-      goto check_node;
-#endif /* not S_IFBLK */
-
-#ifdef S_ISFIFO
-      /* If local system doesn't support FIFOs, use default case.  */
-
     case FIFOTYPE:
-# ifdef S_IFIFO
-      current_stat.st_mode |= S_IFIFO;
-# endif
-      current_stat.st_rdev = 0;	/* FIXME: do we need this? */
-      goto check_node;
-#endif /* S_ISFIFO */
 
-    check_node:
       /* FIXME: deal with umask.  */
 
-      if (!get_stat_data (&stat_data))
+      if (!get_stat_data (current_file_name, &stat_data))
 	break;
 
-      if (current_stat.st_rdev != stat_data.st_rdev)
+      if (current_header->header.typeflag == CHRTYPE
+	  ? !S_ISCHR (stat_data.st_mode)
+	  : current_header->header.typeflag == BLKTYPE
+	  ? !S_ISBLK (stat_data.st_mode)
+	  : /* current_header->header.typeflag == FIFOTYPE */
+	  !S_ISFIFO (stat_data.st_mode))
 	{
-	  report_difference (_("Device numbers changed"));
+	  report_difference (_("File type differs"));
 	  break;
 	}
 
-      if (
-#ifdef S_IFMT
-	  current_stat.st_mode != stat_data.st_mode
-#else
-	  /* POSIX lossage.  */
-	  ((current_stat.st_mode & MODE_ALL)
-	   != (stat_data.st_mode & MODE_ALL))
-#endif
-	  )
+      if ((current_header->header.typeflag == CHRTYPE
+	   || current_header->header.typeflag == BLKTYPE)
+	  && current_stat.st_rdev != stat_data.st_rdev)
 	{
-	  report_difference (_("Mode or device-type changed"));
+	  report_difference (_("Device number differs"));
+	  break;
+	}
+
+      if ((current_stat.st_mode & MODE_ALL) != (stat_data.st_mode & MODE_ALL))
+	{
+	  report_difference (_("Mode differs"));
 	  break;
 	}
 
@@ -705,7 +679,7 @@ diff_archive (void)
 	  read_and_process (current_stat.st_size, process_noop);
 
 	if (multi_volume_option)
-	  assign_string (&save_name, NULL);
+	  assign_string (&save_name, 0);
 	/* Fall through.  */
       }
 
@@ -718,17 +692,21 @@ diff_archive (void)
       while (name_length && current_file_name[name_length] == '/')
 	current_file_name[name_length--] = '\0';	/* zap / */
 
-      if (!get_stat_data (&stat_data))
+      if (!get_stat_data (current_file_name, &stat_data))
 	break;
 
       if (!S_ISDIR (stat_data.st_mode))
 	{
-	  report_difference (_("No longer a directory"));
+	  report_difference (_("File type differs"));
+	  break;
+	}
+
+      if ((current_stat.st_mode & MODE_ALL) != (stat_data.st_mode & MODE_ALL))
+	{
+	  report_difference (_("Mode differs"));
 	  break;
 	}
 
-      if ((stat_data.st_mode & MODE_ALL) != (current_stat.st_mode & MODE_ALL))
-	report_difference (_("Mode differs"));
       break;
 
     case GNUTYPE_VOLHDR:
@@ -742,18 +720,17 @@ diff_archive (void)
 	if (current_file_name[name_length] == '/')
 	  goto really_dir;
 
-	if (!get_stat_data (&stat_data))
+	if (!get_stat_data (current_file_name, &stat_data))
 	  break;
 
 	if (!S_ISREG (stat_data.st_mode))
 	  {
-	    report_difference (_("Not a regular file"));
+	    report_difference (_("File type differs"));
 	    skip_file (current_stat.st_size);
 	    break;
 	  }
 
-	stat_data.st_mode &= MODE_ALL;
-	offset = OFF_FROM_OCT (current_header->oldgnu_header.offset);
+	offset = OFF_FROM_HEADER (current_header->oldgnu_header.offset);
 	if (stat_data.st_size != current_stat.st_size + offset)
 	  {
 	    report_difference (_("Size differs"));
@@ -761,22 +738,20 @@ diff_archive (void)
 	    break;
 	  }
 
-	diff_handle = open (current_file_name, O_NDELAY | O_RDONLY | O_BINARY);
+	diff_handle = open (current_file_name, O_RDONLY | O_BINARY);
 
 	if (diff_handle < 0)
 	  {
-	    WARN ((0, errno, _("Cannot open file %s"), current_file_name));
-	    report_difference (NULL);
+	    open_error (current_file_name);
+	    report_difference (0);
 	    skip_file (current_stat.st_size);
 	    break;
 	  }
 
 	if (lseek (diff_handle, offset, SEEK_SET) < 0)
 	  {
-	    char buf[UINTMAX_STRSIZE_BOUND];
-	    WARN ((0, errno, _("Cannot seek to %s in file %s"),
-		   STRINGIFY_BIGINT (offset, buf), current_file_name));
-	    report_difference (NULL);
+	    seek_error_details (current_file_name, offset);
+	    report_difference (0);
 	    break;
 	  }
 
@@ -790,11 +765,11 @@ diff_archive (void)
 	read_and_process (current_stat.st_size, process_rawdata);
 
 	if (multi_volume_option)
-	  assign_string (&save_name, NULL);
+	  assign_string (&save_name, 0);
 
 	status = close (diff_handle);
-	if (status < 0)
-	  ERROR ((0, errno, _("Error while closing %s"), current_file_name));
+	if (status != 0)
+	  close_error (current_file_name);
 
 	break;
       }
@@ -845,9 +820,7 @@ verify_volume (void)
 	    if (rmtlseek (archive, (off_t) 0, SEEK_SET) != 0)
 	      {
 		/* Lseek failed.  Try a different method.  */
-
-		WARN ((0, errno,
-		       _("Could not rewind archive file for verify")));
+		seek_warn (archive_name_array[0]);
 		return;
 	      }
 #ifdef MTIOCTOP
