@@ -1,5 +1,5 @@
 /* List a tar archive, with support routines for reading a tar archive.
-   Copyright 1988,92,93,94,96,97,98,1999 Free Software Foundation, Inc.
+   Copyright 1988, 92,93,94,96,97,98,99, 2000 Free Software Foundation, Inc.
    Written by John Gilmore, on 1985-08-26.
 
    This program is free software; you can redistribute it and/or modify it
@@ -16,26 +16,50 @@
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-/* Define to non-zero for forcing old ctime() instead of isotime().  */
+/* Define to non-zero for forcing old ctime format instead of ISO format.  */
 #undef USE_OLD_CTIME
 
 #include "system.h"
 #include <quotearg.h>
 
 #include <time.h>
-
-#ifndef FNM_LEADING_DIR
-# include <fnmatch.h>
+#ifndef time
+time_t time ();
 #endif
 
 #include "common.h"
 
+#define max(a, b) ((a) < (b) ? (b) : (a))
+
 union block *current_header;	/* points to current archive header */
 struct stat current_stat;	/* stat struct corresponding */
 enum archive_format current_format; /* recognized format */
 
-static uintmax_t from_oct PARAMS ((const char *, size_t, const char *, uintmax_t));
+static uintmax_t from_header PARAMS ((const char *, size_t, const char *,
+				      uintmax_t, uintmax_t));
 
+/* Base 64 digits; see Internet RFC 2045 Table 1.  */
+static char const base_64_digits[64] =
+{
+  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
+  'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
+  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
+  'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
+  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
+};
+
+/* Table of base-64 digit values indexed by unsigned chars.
+   The value is 64 for unsigned chars that are not base-64 digits.  */
+static char base64_map[UCHAR_MAX + 1];
+
+static void
+base64_init (void)
+{
+  int i;
+  memset (base64_map, 64, sizeof base64_map);
+  for (i = 0; i < 64; i++)
+    base64_map[(int) base_64_digits[i]] = i;
+}
 
 /*-----------------------------------.
 | Main loop for reading an archive.  |
@@ -46,8 +70,8 @@ read_and (void (*do_something) ())
 {
   enum read_header status = HEADER_STILL_UNREAD;
   enum read_header prev_status;
-  char save_typeflag;
 
+  base64_init ();
   name_gather ();
   open_archive (ACCESS_READ);
 
@@ -65,58 +89,45 @@ read_and (void (*do_something) ())
 	  /* Valid header.  We should decode next field (mode) first.
 	     Ensure incoming names are null terminated.  */
 
-	  /* FIXME: This is a quick kludge before 1.12 goes out.  */
-	  current_stat.st_mtime = TIME_FROM_OCT (current_header->header.mtime);
-
-	  if (!name_match (current_file_name)
-	      || current_stat.st_mtime < newer_mtime_option
-	      || excluded_filename (excluded, base_name (current_file_name)))
+	  if (! name_match (current_file_name)
+	      || (newer_mtime_option != TYPE_MINIMUM (time_t)
+		  /* FIXME: We get mtime now, and again later; this causes
+		     duplicate diagnostics if header.mtime is bogus.  */
+		  && ((current_stat.st_mtime
+		       = TIME_FROM_HEADER (current_header->header.mtime))
+		      < newer_mtime_option))
+	      || excluded_name (current_file_name))
 	    {
-	      int isextended = 0;
-
-	      if (current_header->header.typeflag == GNUTYPE_VOLHDR
-		  || current_header->header.typeflag == GNUTYPE_MULTIVOL
-		  || current_header->header.typeflag == GNUTYPE_NAMES)
+	      char save_typeflag = current_header->header.typeflag;
+	      switch (save_typeflag)
 		{
-		  (*do_something) ();
+		case GNUTYPE_VOLHDR:
+		case GNUTYPE_MULTIVOL:
+		case GNUTYPE_NAMES:
+		  break;
+		
+		case DIRTYPE:
+		  if (show_omitted_dirs_option)
+		    WARN ((0, 0, _("%s: Omitting"),
+			   quotearg_colon (current_file_name)));
+		  /* Fall through.  */
+		default:
+
+		  /* Skip past it in the archive.  */
+
+		  set_next_block_after (current_header);
+		  if (current_header->oldgnu_header.isextended)
+		    skip_extended_headers ();
+
+		  /* Skip to the next header on the archive.  */
+
+		  if (save_typeflag != DIRTYPE)
+		    skip_file (current_stat.st_size);
 		  continue;
 		}
-	      if (show_omitted_dirs_option
-		  && current_header->header.typeflag == DIRTYPE)
-		WARN ((0, 0, _("Omitting %s"), current_file_name));
-
-	      /* Skip past it in the archive.  */
-
-	      if (current_header->oldgnu_header.isextended)
-		isextended = 1;
-	      save_typeflag = current_header->header.typeflag;
-	      set_next_block_after (current_header);
-	      if (isextended)
-		{
-#if 0
-		  union block *exhdr;
-
-		  while (1)
-		    {
-		      exhdr = find_next_block ();
-		      if (!exhdr->sparse_header.isextended)
-			{
-			  set_next_block_after (exhdr);
-			  break;
-			}
-		    }
-		  set_next_block_after (exhdr);
-#endif
-		  skip_extended_headers ();
-		}
-
-	      /* Skip to the next header on the archive.  */
-
-	      if (save_typeflag != DIRTYPE)
-		skip_file (current_stat.st_size);
-	      continue;
-	    }
+	      }
 
+	  apply_nonancestor_delayed_set_stat (current_file_name);
 	  (*do_something) ();
 	  continue;
 
@@ -150,12 +161,12 @@ read_and (void (*do_something) ())
 	  switch (prev_status)
 	    {
 	    case HEADER_STILL_UNREAD:
-	      WARN ((0, 0, _("Hmm, this doesn't look like a tar archive")));
+	      ERROR ((0, 0, _("This does not look like a tar archive")));
 	      /* Fall through.  */
 
 	    case HEADER_ZERO_BLOCK:
 	    case HEADER_SUCCESS:
-	      WARN ((0, 0, _("Skipping to next file header")));
+	      ERROR ((0, 0, _("Skipping to next header")));
 	      break;
 
 	    case HEADER_END_OF_FILE:
@@ -180,8 +191,6 @@ read_and (void (*do_something) ())
 void
 list_archive (void)
 {
-  int isextended = 0;		/* to remember if current_header is extended */
-
   /* Print the header block.  */
 
   if (verbose_option)
@@ -210,64 +219,36 @@ list_archive (void)
 	  data_block = find_next_block ();
 	  if (!data_block)
 	    {
-	      ERROR ((0, 0, _("EOF in archive file")));
+	      ERROR ((0, 0, _("Unexpected EOF in archive")));
 	      break;		/* FIXME: What happens, then?  */
 	    }
 	  written = available_space_after (data_block);
 	  if (written > size)
 	    written = size;
-	  errno = 0;		/* FIXME: errno should be read-only */
+	  errno = 0;
 	  check = fwrite (data_block->buffer, sizeof (char), written, stdlis);
 	  set_next_block_after ((union block *)
 				(data_block->buffer + written - 1));
 	  if (check != written)
 	    {
-	      ERROR ((0, errno, _("Only wrote %lu of %lu bytes to file %s"),
-		      (unsigned long) check,
-		      (unsigned long) written, current_file_name));
+	      write_error_details (current_file_name, check, written);
 	      skip_file (size - written);
 	      break;
 	    }
 	}
       if (multi_volume_option)
-	assign_string (&save_name, NULL);
+	assign_string (&save_name, 0);
       fputc ('\n', stdlis);
       fflush (stdlis);
       return;
 
     }
 
-  /* Check to see if we have an extended header to skip over also.  */
-
-  if (current_header->oldgnu_header.isextended)
-    isextended = 1;
-
-  /* Skip past the header in the archive.  */
+  /* Skip past the header in the archive, and past any extended headers.  */
 
   set_next_block_after (current_header);
-
-  /* If we needed to skip any extended headers, do so now, by reading
-     extended headers and skipping past them in the archive.  */
-
-  if (isextended)
-    {
-#if 0
-      union block *exhdr;
-
-      while (1)
-	{
-	  exhdr = find_next_block ();
-
-	  if (!exhdr->sparse_header.isextended)
-	    {
-	      set_next_block_after (exhdr);
-	      break;
-	    }
-	  set_next_block_after (exhdr);
-	}
-#endif
-      skip_extended_headers ();
-    }
+  if (current_header->oldgnu_header.isextended)
+    skip_extended_headers ();
 
   if (multi_volume_option)
     assign_string (&save_name, current_file_name);
@@ -277,7 +258,7 @@ list_archive (void)
   skip_file (current_stat.st_size);
 
   if (multi_volume_option)
-    assign_string (&save_name, NULL);
+    assign_string (&save_name, 0);
 }
 
 /*-----------------------------------------------------------------------.
@@ -296,21 +277,18 @@ list_archive (void)
    bytes in the header as type char.  I think the type char was unsigned
    on the PDP-11, but it's signed on the Next and Sun.  It looks like the
    sources to BSD tar were never changed to compute the checksum
-   currectly, so both the Sun and Next add the bytes of the header as
+   correctly, so both the Sun and Next add the bytes of the header as
    signed chars.  This doesn't cause a problem until you get a file with
    a name containing characters with the high bit set.  So read_header
    computes two checksums -- signed and unsigned.  */
 
-/* FIXME: The signed checksum computation is broken on machines where char's
-   are unsigned.  It's uneasy to handle all cases correctly...  */
-
 enum read_header
 read_header (void)
 {
   size_t i;
-  long unsigned_sum;		/* the POSIX one :-) */
-  long signed_sum;		/* the Sun one :-( */
-  long recorded_sum;
+  int unsigned_sum;		/* the POSIX one :-) */
+  int signed_sum;		/* the Sun one :-( */
+  int recorded_sum;
   uintmax_t parsed_sum;
   char *p;
   union block *header;
@@ -327,42 +305,36 @@ read_header (void)
       if (!header)
 	return HEADER_END_OF_FILE;
 
-      parsed_sum = from_oct (header->header.chksum,
-			     sizeof header->header.chksum,
-			     (char *) 0, TYPE_MAXIMUM (long));
-      if (parsed_sum == (uintmax_t) -1)
-	return HEADER_FAILURE;
-
-      recorded_sum = parsed_sum;
       unsigned_sum = 0;
       signed_sum = 0;
       p = header->buffer;
-      for (i = sizeof (*header); i-- != 0;)
+      for (i = sizeof *header; i-- != 0;)
 	{
-	  /* We can't use unsigned char here because of old compilers,
-	     e.g. V7.  */
-
-	  unsigned_sum += 0xFF & *p;
-	  signed_sum += *p++;
+	  unsigned_sum += (unsigned char) *p;
+	  signed_sum += signed_char (*p++);
 	}
 
+      if (unsigned_sum == 0)
+	return HEADER_ZERO_BLOCK;
+
       /* Adjust checksum to count the "chksum" field as blanks.  */
 
-      for (i = sizeof (header->header.chksum); i-- != 0;)
+      for (i = sizeof header->header.chksum; i-- != 0;)
 	{
-	  unsigned_sum -= 0xFF & header->header.chksum[i];
-	  signed_sum -= header->header.chksum[i];
+	  unsigned_sum -= (unsigned char) header->header.chksum[i];
+	  signed_sum -= signed_char (header->header.chksum[i]);
 	}
       unsigned_sum += ' ' * sizeof header->header.chksum;
       signed_sum += ' ' * sizeof header->header.chksum;
 
-      if (unsigned_sum == sizeof header->header.chksum * ' ')
-	{
-	  /* This is a zeroed block...whole block is 0's except for the
-	     blanks we faked for the checksum field.  */
+      parsed_sum = from_header (header->header.chksum,
+				sizeof header->header.chksum, 0,
+				(uintmax_t) 0,
+				(uintmax_t) TYPE_MAXIMUM (int));
+      if (parsed_sum == (uintmax_t) -1)
+	return HEADER_FAILURE;
 
-	  return HEADER_ZERO_BLOCK;
-	}
+      recorded_sum = parsed_sum;
 
       if (unsigned_sum != recorded_sum && signed_sum != recorded_sum)
 	return HEADER_FAILURE;
@@ -372,9 +344,8 @@ read_header (void)
       if (header->header.typeflag == LNKTYPE)
 	current_stat.st_size = 0;	/* links 0 size on tape */
       else
-	current_stat.st_size = OFF_FROM_OCT (header->header.size);
+	current_stat.st_size = OFF_FROM_HEADER (header->header.size);
 
-      header->header.name[NAME_FIELD_SIZE - 1] = '\0';
       if (header->header.typeflag == GNUTYPE_LONGNAME
 	  || header->header.typeflag == GNUTYPE_LONGLINK)
 	{
@@ -387,15 +358,15 @@ read_header (void)
 	    free (*longp);
 	  size = current_stat.st_size;
 	  if (size != current_stat.st_size)
-	    FATAL_ERROR ((0, 0, _("Memory exhausted")));
-	  bp = *longp = (char *) xmalloc (size);
+	    xalloc_die ();
+	  bp = *longp = xmalloc (size);
 
 	  for (; size > 0; size -= written)
 	    {
 	      data_block = find_next_block ();
-	      if (data_block == NULL)
+	      if (! data_block)
 		{
-		  ERROR ((0, 0, _("Unexpected EOF on archive file")));
+		  ERROR ((0, 0, _("Unexpected EOF in archive")));
 		  break;
 		}
 	      written = available_space_after (data_block);
@@ -413,16 +384,19 @@ read_header (void)
 	}
       else
 	{
-	  char *name = next_long_name;
+	  char *name;
 	  struct posix_header *h = &current_header->header;
-	  char namebuf[sizeof h->prefix + 1 + sizeof h->name + 1];
+	  char namebuf[sizeof h->prefix + 1 + NAME_FIELD_SIZE + 1];
 
+	  name = next_long_name;
 	  if (! name)
 	    {
 	      /* Accept file names as specified by POSIX.1-1996
                  section 10.1.1.  */
+	      int posix_header = strcmp (h->magic, TMAGIC) == 0;
 	      char *np = namebuf;
-	      if (h->prefix[0])
+
+	      if (posix_header && h->prefix[0])
 		{
 		  memcpy (np, h->prefix, sizeof h->prefix);
 		  np[sizeof h->prefix] = '\0';
@@ -433,12 +407,27 @@ read_header (void)
 	      np[sizeof h->name] = '\0';
 	      name = namebuf;
 	    }
-
 	  assign_string (&current_file_name, name);
-	  assign_string (&current_link_name,
-			 (next_long_link ? next_long_link
-			  : current_header->header.linkname));
-	  next_long_link = next_long_name = 0;
+	  if (next_long_name)
+	    {
+	      free (next_long_name);
+	      next_long_name = 0;
+	    }
+	  
+	  name = next_long_link;
+	  if (! name)
+	    {
+	      memcpy (namebuf, h->linkname, sizeof h->linkname);
+	      namebuf[sizeof h->linkname] = '\0';
+	      name = namebuf;
+	    }
+	  assign_string (&current_link_name, name);
+	  if (next_long_link)
+	    {
+	      free (next_long_link);
+	      next_long_link = 0;
+	    }
+
 	  return HEADER_SUCCESS;
 	}
     }
@@ -474,19 +463,19 @@ decode_header (union block *header, struct stat *stat_info,
     format = V7_FORMAT;
   *format_pointer = format;
 
-  stat_info->st_mode = MODE_FROM_OCT (header->header.mode);
-  stat_info->st_mtime = TIME_FROM_OCT (header->header.mtime);
+  stat_info->st_mode = MODE_FROM_HEADER (header->header.mode);
+  stat_info->st_mtime = TIME_FROM_HEADER (header->header.mtime);
 
   if (format == OLDGNU_FORMAT && incremental_option)
     {
-      stat_info->st_atime = TIME_FROM_OCT (header->oldgnu_header.atime);
-      stat_info->st_ctime = TIME_FROM_OCT (header->oldgnu_header.ctime);
+      stat_info->st_atime = TIME_FROM_HEADER (header->oldgnu_header.atime);
+      stat_info->st_ctime = TIME_FROM_HEADER (header->oldgnu_header.ctime);
     }
 
   if (format == V7_FORMAT)
     {
-      stat_info->st_uid = UID_FROM_OCT (header->header.uid);
-      stat_info->st_gid = GID_FROM_OCT (header->header.gid);
+      stat_info->st_uid = UID_FROM_HEADER (header->header.uid);
+      stat_info->st_gid = GID_FROM_HEADER (header->header.gid);
       stat_info->st_rdev = 0;
     }
   else
@@ -498,30 +487,26 @@ decode_header (union block *header, struct stat *stat_info,
 	  if (numeric_owner_option
 	      || !*header->header.uname
 	      || !uname_to_uid (header->header.uname, &stat_info->st_uid))
-	    stat_info->st_uid = UID_FROM_OCT (header->header.uid);
+	    stat_info->st_uid = UID_FROM_HEADER (header->header.uid);
 
 	  if (numeric_owner_option
 	      || !*header->header.gname
 	      || !gname_to_gid (header->header.gname, &stat_info->st_gid))
-	    stat_info->st_gid = GID_FROM_OCT (header->header.gid);
+	    stat_info->st_gid = GID_FROM_HEADER (header->header.gid);
 	}
       switch (header->header.typeflag)
 	{
-#ifdef S_IFBLK
 	case BLKTYPE:
 	  stat_info->st_rdev
-	    = makedev (MAJOR_FROM_OCT (header->header.devmajor),
-		       MINOR_FROM_OCT (header->header.devminor));
+	    = makedev (MAJOR_FROM_HEADER (header->header.devmajor),
+		       MINOR_FROM_HEADER (header->header.devminor));
 	  break;
-#endif
 
-#ifdef S_IFCHR
 	case CHRTYPE:
 	  stat_info->st_rdev
-	    = makedev (MAJOR_FROM_OCT (header->header.devmajor),
-		       MINOR_FROM_OCT (header->header.devminor));
+	    = makedev (MAJOR_FROM_HEADER (header->header.devmajor),
+		       MINOR_FROM_HEADER (header->header.devminor));
 	  break;
-#endif
 
 	default:
 	  stat_info->st_rdev = 0;
@@ -530,44 +515,165 @@ decode_header (union block *header, struct stat *stat_info,
 }
 
 /*------------------------------------------------------------------------.
-| Quick and dirty octal conversion.  Result is -1 if the field is invalid |
-| (all blank, or nonoctal).						  |
+| Convert buffer at WHERE0 of size DIGS from external format to uintmax_t.|
+| The data is of type TYPE.  The buffer must represent a value in the     |
+| range -MINUS_MINVAL through MAXVAL.  DIGS must be positive.		  |
 `------------------------------------------------------------------------*/
 
 static uintmax_t
-from_oct (const char *where0, size_t digs0, const char *type, uintmax_t maxval)
+from_header (char const *where0, size_t digs, char const *type,
+	     uintmax_t minus_minval, uintmax_t maxval)
 {
   uintmax_t value;
-  const char *where = where0;
-  size_t digs = digs0;
+  char const *where = where0;
+  char const *lim = where + digs;
+  int negative = 0;
 
+  /* Accommodate buggy tar of unknown vintage, which outputs leading
+     NUL if the previous field overflows.  */
+  where += !*where;
+
+  /* Accommodate older tars, which output leading spaces.  */
   for (;;)
     {
-      if (digs == 0)
+      if (where == lim)
 	{
 	  if (type)
-	    ERROR ((0, 0, _("Blanks in header where octal %s value expected"),
+	    ERROR ((0, 0,
+		    _("Blanks in header where numeric %s value expected"),
 		    type));
 	  return -1;
 	}
       if (!ISSPACE ((unsigned char) *where))
 	break;
       where++;
-      digs--;
     }
 
   value = 0;
-  while (digs != 0 && ISODIGIT (*where))
+  if (ISODIGIT (*where))
     {
-      /* Scan til nonoctal.  */
+      char const *where1 = where;
+      uintmax_t overflow = 0;
 
-      if (value << 3 >> 3 != value)
-	goto out_of_range;
-      value = (value << 3) | (*where++ - '0');
-      --digs;
+      for (;;)
+	{
+	  value += *where++ - '0';
+	  if (where == lim || ! ISODIGIT (*where))
+	    break;
+	  overflow |= value ^ (value << LG_8 >> LG_8);
+	  value <<= LG_8;
+	}
+
+      /* Parse the output of older, unportable tars, which generate
+         negative values in two's complement octal.  If the leading
+         nonzero digit is 1, we can't recover the original value
+         reliably; so do this only if the digit is 2 or more.  This
+         catches the common case of 32-bit negative time stamps.  */
+      if ((overflow || maxval < value) && '2' <= *where1)
+	{
+	  /* Compute the negative of the input value, assuming two's
+	     complement.  */
+	  int digit = (*where1 - '0') | 4;
+	  overflow = 0;
+	  value = 0;
+	  where = where1;
+	  for (;;)
+	    {
+	      value += 7 - digit;
+	      where++;
+	      if (where == lim || ! ISODIGIT (*where))
+		break;
+	      digit = *where - '0';
+	      overflow |= value ^ (value << LG_8 >> LG_8);
+	      value <<= LG_8;
+	    }
+	  value++;
+	  overflow |= !value;
+
+	  if (!overflow && value <= minus_minval)
+	    {
+	      WARN ((0, 0,
+		     _("Archive octal value %.*s is out of %s range; assuming two's complement"),
+		     (int) (where - where1), where1, type));
+	      negative = 1;
+	    }
+	}
+
+      if (overflow)
+	{
+	  ERROR ((0, 0,
+		  _("Archive octal value %.*s is out of %s range"),
+		  (int) (where - where1), where1, type));
+	  return -1;
+	}
+    }
+  else if (type)
+    {
+      /* The following forms cannot appear as checksums, so we don't
+	 check for them if TYPE is null.  */
+
+      if (*where == '-' || *where == '+')
+	{
+	  /* Parse base-64 output produced only by tar test versions
+	     1.13.6 (1999-08-11) through 1.13.11 (1999-08-23).
+	     Support for this will be withdrawn in future releases.  */
+	  int dig;
+	  static int warned_once;
+	  if (! warned_once)
+	    {
+	      warned_once = 1;
+	      WARN ((0, 0,
+		     _("Archive contains obsolescent base-64 headers")));
+	    }
+	  negative = *where++ == '-';
+	  while (where != lim
+		 && (dig = base64_map[(unsigned char) *where]) < 64)
+	    {
+	      if (value << LG_64 >> LG_64 != value)
+		{
+		  ERROR ((0, 0,
+			  _("Archive signed base-64 string `%.*s' is out of %s range"),
+			  (int) digs, where0, type));
+		  return -1;
+		}
+	      value = (value << LG_64) | dig;
+	      where++;
+	    }
+	}
+      else if (*where == '\200' /* positive base-256 */
+	       || *where == '\377' /* negative base-256 */)
+	{
+	  /* Parse base-256 output.  A nonnegative number N is
+	     represented as (256**DIGS)/2 + N; a negative number -N is
+	     represented as (256**DIGS) - N, i.e. as two's complement.
+	     The representation guarantees that the leading bit is
+	     always on, so that we don't confuse this format with the
+	     others (assuming ASCII bytes of 8 bits or more).  */
+	  int signbit = *where & (1 << (LG_256 - 2));
+	  uintmax_t topbits = (((uintmax_t) - signbit)
+			       << (CHAR_BIT * sizeof (uintmax_t)
+				   - LG_256 - (LG_256 - 2)));
+	  value = (*where++ & ((1 << (LG_256 - 2)) - 1)) - signbit;
+	  for (;;)
+	    {
+	      value = (value << LG_256) + (unsigned char) *where++;
+	      if (where == lim)
+		break;
+	      if (((value << LG_256 >> LG_256) | topbits) != value)
+		{
+		  ERROR ((0, 0,
+			  _("Archive base-256 value is out of %s range"),
+			  type));
+		  return -1;
+		}
+	    }
+	  negative = signbit;
+	  if (negative)
+	    value = -value;
+	}
     }
 
-  if (digs != 0 && *where && !ISSPACE ((unsigned char) *where))
+  if (where != lim && *where && !ISSPACE ((unsigned char) *where))
     {
       if (type)
 	{
@@ -576,48 +682,74 @@ from_oct (const char *where0, size_t digs0, const char *type, uintmax_t maxval)
 
 	  if (!o)
 	    {
-	      o = clone_quoting_options ((struct quoting_options *) 0);
-	      set_quoting_style (o, escape_quoting_style);
+	      o = clone_quoting_options (0);
+	      set_quoting_style (o, locale_quoting_style);
 	    }
 
-	  quotearg_buffer (buf, sizeof buf, where0, digs0, o);
+	  while (where0 != lim && ! lim[-1])
+	    lim--;
+	  quotearg_buffer (buf, sizeof buf, where0, lim - where, o);
 	  ERROR ((0, 0,
-		  _("Header contains \"%.*s\" where octal %s value expected"),
+		  _("Archive contains %.*s where numeric %s value expected"),
 		  (int) sizeof buf, buf, type));
 	}
 
       return -1;
     }
 
-  if (value <= maxval)
-    return value;
+  if (value <= (negative ? minus_minval : maxval))
+    return negative ? -value : value;
 
- out_of_range:
   if (type)
-    ERROR ((0, 0, _("Octal value `%.*s' is out of range for %s"),
-	    (int) digs0, where0, type));
+    {
+      char minval_buf[UINTMAX_STRSIZE_BOUND + 1];
+      char maxval_buf[UINTMAX_STRSIZE_BOUND];
+      char value_buf[UINTMAX_STRSIZE_BOUND + 1];
+      char *minval_string = STRINGIFY_BIGINT (minus_minval, minval_buf + 1);
+      char *value_string = STRINGIFY_BIGINT (value, value_buf + 1);
+      if (negative)
+	*--value_string = '-';
+      if (minus_minval)
+	*--minval_string = '-';
+      ERROR ((0, 0, _("Archive value %s is out of %s range %s..%s"),
+	      value_string, type,
+	      minval_string, STRINGIFY_BIGINT (maxval, maxval_buf)));
+    }
+
   return -1;
 }
+
 gid_t
-gid_from_oct (const char *p, size_t s)
+gid_from_header (const char *p, size_t s)
 {
-  return from_oct (p, s, "gid_t", (uintmax_t) TYPE_MAXIMUM (gid_t));
+  return from_header (p, s, "gid_t",
+		      - (uintmax_t) TYPE_MINIMUM (gid_t),
+		      (uintmax_t) TYPE_MAXIMUM (gid_t));
 }
+
 major_t
-major_from_oct (const char *p, size_t s)
+major_from_header (const char *p, size_t s)
 {
-  return from_oct (p, s, "major_t", (uintmax_t) TYPE_MAXIMUM (major_t));
+  return from_header (p, s, "major_t",
+		      - (uintmax_t) TYPE_MINIMUM (major_t),
+		      (uintmax_t) TYPE_MAXIMUM (major_t));
 }
+
 minor_t
-minor_from_oct (const char *p, size_t s)
+minor_from_header (const char *p, size_t s)
 {
-  return from_oct (p, s, "minor_t", (uintmax_t) TYPE_MAXIMUM (minor_t));
+  return from_header (p, s, "minor_t",
+		      - (uintmax_t) TYPE_MINIMUM (minor_t),
+		      (uintmax_t) TYPE_MAXIMUM (minor_t));
 }
+
 mode_t
-mode_from_oct (const char *p, size_t s)
+mode_from_header (const char *p, size_t s)
 {
   /* Do not complain about unrecognized mode bits.  */
-  unsigned u = from_oct (p, s, "mode_t", TYPE_MAXIMUM (uintmax_t));
+  unsigned u = from_header (p, s, "mode_t",
+			    - (uintmax_t) TYPE_MINIMUM (mode_t),
+			    TYPE_MAXIMUM (uintmax_t));
   return ((u & TSUID ? S_ISUID : 0)
 	  | (u & TSGID ? S_ISGID : 0)
 	  | (u & TSVTX ? S_ISVTX : 0)
@@ -631,32 +763,48 @@ mode_from_oct (const char *p, size_t s)
 	  | (u & TOWRITE ? S_IWOTH : 0)
 	  | (u & TOEXEC ? S_IXOTH : 0));
 }
+
 off_t
-off_from_oct (const char *p, size_t s)
+off_from_header (const char *p, size_t s)
 {
-  return from_oct (p, s, "off_t", (uintmax_t) TYPE_MAXIMUM (off_t));
-}
-size_t
-size_from_oct (const char *p, size_t s)
-{
-  return from_oct (p, s, "size_t", (uintmax_t) TYPE_MAXIMUM (size_t));
-}
-time_t
-time_from_oct (const char *p, size_t s)
-{
-  return from_oct (p, s, "time_t", (uintmax_t) TYPE_MAXIMUM (time_t));
-}
-uid_t
-uid_from_oct (const char *p, size_t s)
-{
-  return from_oct (p, s, "uid_t", (uintmax_t) TYPE_MAXIMUM (uid_t));
-}
-uintmax_t
-uintmax_from_oct (const char *p, size_t s)
-{
-  return from_oct (p, s, "uintmax_t", TYPE_MAXIMUM (uintmax_t));
+  /* Negative offsets are not allowed in tar files, so invoke
+     from_header with minimum value 0, not TYPE_MINIMUM (off_t).  */
+  return from_header (p, s, "off_t", (uintmax_t) 0,
+		      (uintmax_t) TYPE_MAXIMUM (off_t));
 }
 
+size_t
+size_from_header (const char *p, size_t s)
+{
+  return from_header (p, s, "size_t", (uintmax_t) 0, 
+		      (uintmax_t) TYPE_MAXIMUM (size_t));
+}
+
+time_t
+time_from_header (const char *p, size_t s)
+{
+  time_t t = from_header (p, s, "time_t",
+			  - (uintmax_t) TYPE_MINIMUM (time_t),
+			  (uintmax_t) TYPE_MAXIMUM (time_t));
+  if (start_time < t && time (0) < t)
+    WARN ((0, 0, _("Archive contains future timestamp %s"), tartime (t)));
+  return t;
+}
+
+uid_t
+uid_from_header (const char *p, size_t s)
+{
+  return from_header (p, s, "uid_t",
+		      - (uintmax_t) TYPE_MINIMUM (uid_t),
+		      (uintmax_t) TYPE_MAXIMUM (uid_t));
+}
+
+uintmax_t
+uintmax_from_header (const char *p, size_t s)
+{
+  return from_header (p, s, "uintmax_t", (uintmax_t) 0,
+		      TYPE_MAXIMUM (uintmax_t));
+}
 
 
 /*----------------------------------------------------------------------.
@@ -673,53 +821,50 @@ stringify_uintmax_t_backwards (uintmax_t o, char *buf)
   return buf;
 }
 
-#if !USE_OLD_CTIME
-
-/*-------------------------------------------.
-| Return the time formatted along ISO 8601.  |
-`-------------------------------------------*/
-
-/* Also, see http://www.ft.uni-erlangen.de/~mskuhn/iso-time.html.  */
-
-static char *
-isotime (const time_t *time)
+/* Return a printable representation of T.  The result points to
+   static storage that can be reused in the next call to this
+   function, to ctime, or to asctime.  */
+char const *
+tartime (time_t t)
 {
-  static char buffer[21];
-  struct tm *tm;
+  static char buffer[max (UINTMAX_STRSIZE_BOUND + 1,
+			  INT_STRLEN_BOUND (int) + 16)];
+  char *p;
 
-  tm = localtime (time);
-  sprintf (buffer, "%4d-%02d-%02d %02d:%02d:%02d\n",
-	   tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
-	   tm->tm_hour, tm->tm_min, tm->tm_sec);
-  return buffer;
-}
+#if USE_OLD_CTIME
+  p = ctime (&t);
+  if (p)
+    {
+      char const *time_stamp = p + 4;
+      for (p += 16; p[4] != '\n'; p++)
+	p[0] = p[4];
+      p[0] = '\0';
+      return time_stamp;
+    }
+#else
+  /* Use ISO 8610 format.  See:
+     http://www.cl.cam.ac.uk/~mgk25/iso-time.html  */
+  struct tm *tm = localtime (&t);
+  if (tm)
+    {
+      sprintf (buffer, "%04d-%02d-%02d %02d:%02d:%02d",
+	       tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
+	       tm->tm_hour, tm->tm_min, tm->tm_sec);
+      return buffer;
+    }
+#endif
 
-#endif /* not USE_OLD_CTIME */
-
-/*-------------------------------------------------------------------------.
-| Decode MODE from its binary form in a stat structure, and encode it into |
-| a 9 characters string STRING, terminated with a NUL.                     |
-`-------------------------------------------------------------------------*/
-
-static void
-decode_mode (mode_t mode, char *string)
-{
-  *string++ = mode & S_IRUSR ? 'r' : '-';
-  *string++ = mode & S_IWUSR ? 'w' : '-';
-  *string++ = (mode & S_ISUID
-	       ? (mode & S_IXUSR ? 's' : 'S')
-	       : (mode & S_IXUSR ? 'x' : '-'));
-  *string++ = mode & S_IRGRP ? 'r' : '-';
-  *string++ = mode & S_IWGRP ? 'w' : '-';
-  *string++ = (mode & S_ISGID
-	       ? (mode & S_IXGRP ? 's' : 'S')
-	       : (mode & S_IXGRP ? 'x' : '-'));
-  *string++ = mode & S_IROTH ? 'r' : '-';
-  *string++ = mode & S_IWOTH ? 'w' : '-';
-  *string++ = (mode & S_ISVTX
-	       ? (mode & S_IXOTH ? 't' : 'T')
-	       : (mode & S_IXOTH ? 'x' : '-'));
-  *string = '\0';
+  /* The time stamp cannot be broken down, most likely because it
+     is out of range.  Convert it as an integer,
+     right-adjusted in a field with the same width as the usual
+     19-byte 4-year ISO time format.  */
+  p = stringify_uintmax_t_backwards (t < 0 ? - (uintmax_t) t : (uintmax_t) t,
+				     buffer + sizeof buffer);
+  if (t < 0)
+    *--p = '-';
+  while (buffer + sizeof buffer - 19 - 1 < p)
+    *--p = ' ';
+  return p;
 }
 
 /*-------------------------------------------------------------------------.
@@ -752,16 +897,14 @@ void
 print_header (void)
 {
   char modes[11];
-  char *timestamp;
+  char const *time_stamp;
   /* These hold formatted ints.  */
   char uform[UINTMAX_STRSIZE_BOUND], gform[UINTMAX_STRSIZE_BOUND];
   char *user, *group;
   char size[2 * UINTMAX_STRSIZE_BOUND];
   				/* holds formatted size or major,minor */
   char uintbuf[UINTMAX_STRSIZE_BOUND];
-  time_t longie;		/* to make ctime() call portable */
   int pad;
-  char *name;
 
   if (block_number_option)
     {
@@ -773,16 +916,7 @@ print_header (void)
   if (verbose_option <= 1)
     {
       /* Just the fax, mam.  */
-
-      char *quoted_name = quote_copy_string (current_file_name);
-
-      if (quoted_name)
-	{
-	  fprintf (stdlis, "%s\n", quoted_name);
-	  free (quoted_name);
-	}
-      else
-	fprintf (stdlis, "%s\n", current_file_name);
+      fprintf (stdlis, "%s\n", quotearg (current_file_name));
     }
   else
     {
@@ -841,30 +975,25 @@ print_header (void)
 
       decode_mode (current_stat.st_mode, modes + 1);
 
-      /* Timestamp.  */
+      /* Time stamp.  */
 
-      longie = current_stat.st_mtime;
-#if USE_OLD_CTIME
-      timestamp = ctime (&longie);
-      timestamp[16] = '\0';
-      timestamp[24] = '\0';
-#else
-      timestamp = isotime (&longie);
-      timestamp[16] = '\0';
-#endif
+      time_stamp = tartime (current_stat.st_mtime);
 
       /* User and group names.  */
 
-      if (*current_header->header.uname && current_format != V7_FORMAT)
+      if (*current_header->header.uname && current_format != V7_FORMAT
+	  && !numeric_owner_option)
 	user = current_header->header.uname;
       else
-	user = STRINGIFY_BIGINT (UINTMAX_FROM_OCT (current_header->header.uid),
+	user = STRINGIFY_BIGINT (UINTMAX_FROM_HEADER
+				 (current_header->header.uid),
 				 uform);
 
-      if (*current_header->header.gname && current_format != V7_FORMAT)
+      if (*current_header->header.gname && current_format != V7_FORMAT
+	  && !numeric_owner_option)
 	group = current_header->header.gname;
       else
-	group = STRINGIFY_BIGINT (UINTMAX_FROM_OCT
+	group = STRINGIFY_BIGINT (UINTMAX_FROM_HEADER
 				  (current_header->header.gid),
 				  gform);
 
@@ -872,18 +1001,19 @@ print_header (void)
 
       switch (current_header->header.typeflag)
 	{
-#if defined(S_IFBLK) || defined(S_IFCHR)
 	case CHRTYPE:
 	case BLKTYPE:
-	  sprintf (size, "%lu,%lu",
-		   (unsigned long) major (current_stat.st_rdev),
-		   (unsigned long) minor (current_stat.st_rdev));
+	  strcpy (size,
+		  STRINGIFY_BIGINT (major (current_stat.st_rdev), uintbuf));
+	  strcat (size, ",");
+	  strcat (size,
+		  STRINGIFY_BIGINT (minor (current_stat.st_rdev), uintbuf));
 	  break;
-#endif
 	case GNUTYPE_SPARSE:
 	  strcpy (size,
 		  STRINGIFY_BIGINT
-		  (UINTMAX_FROM_OCT (current_header->oldgnu_header.realsize),
+		  (UINTMAX_FROM_HEADER (current_header
+					->oldgnu_header.realsize),
 		   uintbuf));
 	  break;
 	default:
@@ -897,46 +1027,19 @@ print_header (void)
       if (pad > ugswidth)
 	ugswidth = pad;
 
-#if USE_OLD_CTIME
-      fprintf (stdlis, "%s %s/%s %*s%s %s %s",
-	       modes, user, group, ugswidth - pad, "",
-	       size, timestamp + 4, timestamp + 20);
-#else
       fprintf (stdlis, "%s %s/%s %*s%s %s",
-	       modes, user, group, ugswidth - pad, "", size, timestamp);
-#endif
+	       modes, user, group, ugswidth - pad, "", size, time_stamp);
 
-      name = quote_copy_string (current_file_name);
-      if (name)
-	{
-	  fprintf (stdlis, " %s", name);
-	  free (name);
-	}
-      else
-	fprintf (stdlis, " %s", current_file_name);
+      fprintf (stdlis, " %s", quotearg (current_file_name));
 
       switch (current_header->header.typeflag)
 	{
 	case SYMTYPE:
-	  name = quote_copy_string (current_link_name);
-	  if (name)
-	    {
-	      fprintf (stdlis, " -> %s\n", name);
-	      free (name);
-	    }
-	  else
-	    fprintf (stdlis, " -> %s\n", current_link_name);
+	  fprintf (stdlis, " -> %s\n", quotearg (current_link_name));
 	  break;
 
 	case LNKTYPE:
-	  name = quote_copy_string (current_link_name);
-	  if (name)
-	    {
-	      fprintf (stdlis, _(" link to %s\n"), name);
-	      free (name);
-	    }
-	  else
-	    fprintf (stdlis, _(" link to %s\n"), current_link_name);
+	  fprintf (stdlis, _(" link to %s\n"), quotearg (current_link_name));
 	  break;
 
 	default:
@@ -963,7 +1066,7 @@ print_header (void)
 	case GNUTYPE_MULTIVOL:
 	  strcpy (size,
 		  STRINGIFY_BIGINT
-		  (UINTMAX_FROM_OCT (current_header->oldgnu_header.offset),
+		  (UINTMAX_FROM_HEADER (current_header->oldgnu_header.offset),
 		   uintbuf));
 	  fprintf (stdlis, _("--Continued at byte %s--\n"), size);
 	  break;
@@ -984,7 +1087,6 @@ void
 print_for_mkdir (char *pathname, int length, mode_t mode)
 {
   char modes[11];
-  char *name;
 
   if (verbose_option > 1)
     {
@@ -999,16 +1101,9 @@ print_for_mkdir (char *pathname, int length, mode_t mode)
 	  fprintf (stdlis, _("block %s: "),
 		   STRINGIFY_BIGINT (current_block_ordinal (), buf));
 	}
-      name = quote_copy_string (pathname);
-      if (name)
-	{
-	  fprintf (stdlis, "%s %*s %.*s\n", modes, ugswidth + DATEWIDTH,
-		   _("Creating directory:"), length, name);
-	  free (name);
-	}
-      else
-	fprintf (stdlis, "%s %*s %.*s\n", modes, ugswidth + DATEWIDTH,
-		 _("Creating directory:"), length, pathname);
+
+      fprintf (stdlis, "%s %*s %.*s\n", modes, ugswidth + DATEWIDTH,
+	       _("Creating directory:"), length, quotearg (pathname));
     }
 }
 
@@ -1030,8 +1125,8 @@ skip_file (off_t size)
   while (size > 0)
     {
       x = find_next_block ();
-      if (x == NULL)
-	FATAL_ERROR ((0, 0, _("Unexpected EOF on archive file")));
+      if (! x)
+	FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
 
       set_next_block_after (x);
       size -= BLOCKSIZE;
@@ -1049,14 +1144,12 @@ skip_extended_headers (void)
 {
   union block *exhdr;
 
-  while (1)
+  do
     {
       exhdr = find_next_block ();
-      if (!exhdr->sparse_header.isextended)
-	{
-	  set_next_block_after (exhdr);
-	  break;
-	}
+      if (!exhdr)
+	FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
       set_next_block_after (exhdr);
     }
+  while (exhdr->sparse_header.isextended);
 }
