@@ -15,14 +15,17 @@
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
+/* Enable GNU extensions in fnmatch.h.  */
+#ifndef _GNU_SOURCE
+# define _GNU_SOURCE 1
+#endif
+
 #include "system.h"
 
 #include <pwd.h>
 #include <grp.h>
-
-#ifndef FNM_LEADING_DIR
-# include <fnmatch.h>
-#endif
+#include <fnmatch.h>
+#include <quotearg.h>
 
 #include "common.h"
 
@@ -42,20 +45,20 @@ extern struct group *getgrgid ();
    This code should also be modified for non-UNIX systems to do something
    reasonable.  */
 
-static char cached_uname[UNAME_FIELD_SIZE] = "";
-static char cached_gname[GNAME_FIELD_SIZE] = "";
+static char cached_uname[UNAME_FIELD_SIZE];
+static char cached_gname[GNAME_FIELD_SIZE];
 
 static uid_t cached_uid;	/* valid only if cached_uname is not empty */
 static gid_t cached_gid;	/* valid only if cached_gname is not empty */
 
 /* These variables are valid only if nonempty.  */
-static char cached_no_such_uname[UNAME_FIELD_SIZE] = "";
-static char cached_no_such_gname[GNAME_FIELD_SIZE] = "";
+static char cached_no_such_uname[UNAME_FIELD_SIZE];
+static char cached_no_such_gname[GNAME_FIELD_SIZE];
 
 /* These variables are valid only if nonzero.  It's not worth optimizing
    the case for weird systems where 0 is not a valid uid or gid.  */
-static uid_t cached_no_such_uid = 0;
-static gid_t cached_no_such_gid = 0;
+static uid_t cached_no_such_uid;
+static gid_t cached_no_such_gid;
 
 /*------------------------------------------.
 | Given UID, find the corresponding UNAME.  |
@@ -195,7 +198,7 @@ gname_to_gid (char gname[GNAME_FIELD_SIZE], gid_t *gidp)
 static const char **name_array;	/* store an array of names */
 static int allocated_names;	/* how big is the array? */
 static int names;		/* how many entries does it have? */
-static int name_index = 0;	/* how many of the entries have we scanned? */
+static int name_index;		/* how many of the entries have we scanned? */
 
 /*------------------------.
 | Initialize structures.  |
@@ -205,8 +208,7 @@ void
 init_names (void)
 {
   allocated_names = 10;
-  name_array = (const char **)
-    xmalloc (sizeof (const char *) * allocated_names);
+  name_array = xmalloc (sizeof (const char *) * allocated_names);
   names = 0;
 }
 
@@ -220,7 +222,7 @@ name_add (const char *name)
   if (names == allocated_names)
     {
       allocated_names *= 2;
-      name_array = (const char **)
+      name_array =
 	xrealloc (name_array, sizeof (const char *) * allocated_names);
     }
   name_array[names++] = name;
@@ -265,7 +267,7 @@ name_init (int argc, char *const *argv)
 	  name_file = stdin;
 	}
       else if (name_file = fopen (files_from_option, "r"), !name_file)
-	FATAL_ERROR ((0, errno, _("Cannot open file %s"), files_from_option));
+	open_fatal (files_from_option);
     }
 }
 
@@ -343,16 +345,15 @@ name_next (int change_dirs)
     {
       /* Get a name, either from file or from saved arguments.  */
 
-      if (name_file)
+      if (name_index == names)
 	{
-	  if (!read_name_from_file ())
+	  if (! name_file)
+	    break;
+	  if (! read_name_from_file ())
 	    break;
 	}
       else
 	{
-	  if (name_index == names)
-	    break;
-
 	  source = name_array[name_index++];
 	  if (strlen (source) > name_buffer_length)
 	    {
@@ -372,8 +373,7 @@ name_next (int change_dirs)
       if (chdir_flag)
 	{
 	  if (chdir (name_buffer) < 0)
-	    FATAL_ERROR ((0, errno, _("Cannot change to directory %s"),
-			  name_buffer));
+	    chdir_fatal (name_buffer);
 	  chdir_flag = 0;
 	}
       else if (change_dirs && strcmp (name_buffer, "-C") == 0)
@@ -390,7 +390,7 @@ name_next (int change_dirs)
   if (name_file && chdir_flag)
     FATAL_ERROR ((0, 0, _("Missing file name after -C")));
 
-  return NULL;
+  return 0;
 }
 
 /*------------------------------.
@@ -400,9 +400,9 @@ name_next (int change_dirs)
 void
 name_close (void)
 {
-  if (name_file != NULL && name_file != stdin)
-    if (fclose (name_file) == EOF)
-      ERROR ((0, errno, "%s", name_buffer));
+  if (name_file && name_file != stdin)
+    if (fclose (name_file) != 0)
+      close_error (name_buffer);
 }
 
 /*-------------------------------------------------------------------------.
@@ -422,53 +422,74 @@ name_gather (void)
 {
   /* Buffer able to hold a single name.  */
   static struct name *buffer;
-  static size_t allocated_length = 0;
+  static size_t allocated_length;
 
-  char *name;
+  char const *name;
 
   if (same_order_option)
     {
+      static int change_dir;
+
       if (allocated_length == 0)
 	{
 	  allocated_length = sizeof (struct name) + NAME_FIELD_SIZE;
-	  buffer = (struct name *) xmalloc (allocated_length);
+	  buffer = xmalloc (allocated_length);
 	  /* FIXME: This memset is overkill, and ugly...  */
 	  memset (buffer, 0, allocated_length);
 	}
-      name = name_next (0);
+
+      while ((name = name_next (0)) && strcmp (name, "-C") == 0)
+	{
+	  char const *dir = name_next (0);
+	  if (! dir)
+	    FATAL_ERROR ((0, 0, _("Missing file name after -C")));
+	  change_dir = chdir_arg (xstrdup (dir));
+	}
+
       if (name)
 	{
-	  if (strcmp (name, "-C") == 0)
-	    {
-	      char *copy = xstrdup (name_next (0));
-
-	      name = name_next (0);
-	      if (!name)
-		FATAL_ERROR ((0, 0, _("Missing file name after -C")));
-	      buffer->change_dir = copy;
-	    }
 	  buffer->length = strlen (name);
 	  if (sizeof (struct name) + buffer->length >= allocated_length)
 	    {
 	      allocated_length = sizeof (struct name) + buffer->length;
-	      buffer = (struct name *) xrealloc (buffer, allocated_length);
+	      buffer = xrealloc (buffer, allocated_length);
 	    }
-	  strncpy (buffer->name, name, (size_t) buffer->length);
+	  buffer->change_dir = change_dir;
+	  strncpy (buffer->name, name, buffer->length);
 	  buffer->name[buffer->length] = 0;
-	  buffer->next = NULL;
+	  buffer->next = 0;
 	  buffer->found = 0;
 
 	  /* FIXME: Poorly named globals, indeed...  */
 	  namelist = buffer;
 	  namelast = namelist;
 	}
-      return;
     }
+  else
+    {
+      /* Non sorted names -- read them all in.  */
+      int change_dir = 0;
 
-  /* Non sorted names -- read them all in.  */
-
-  while (name = name_next (0), name)
-    addname (name);
+      for (;;)
+	{
+	  int change_dir0 = change_dir;
+	  while ((name = name_next (0)) && strcmp (name, "-C") == 0)
+	    {
+	      char const *dir = name_next (0);
+	      if (! dir)
+		FATAL_ERROR ((0, 0, _("Missing file name after -C")));
+	      change_dir = chdir_arg (xstrdup (dir));
+	    }
+	  if (name)
+	    addname (name, change_dir);
+	  else
+	    {
+	      if (change_dir != change_dir0)
+		addname (0, change_dir);
+	      break;
+	    }
+	}
+    }
 }
 
 /*-----------------------------.
@@ -476,53 +497,21 @@ name_gather (void)
 `-----------------------------*/
 
 void
-addname (const char *string)
+addname (char const *string, int change_dir)
 {
-  /* FIXME: This is ugly.  How is memory managed?  */
-  static char *chdir_name = NULL;
-
   struct name *name;
   size_t length;
 
-  if (strcmp (string, "-C") == 0)
-    {
-      chdir_name = xstrdup (name_next (0));
-      string = name_next (0);
-      if (!chdir_name)
-	FATAL_ERROR ((0, 0, _("Missing file name after -C")));
-
-      if (chdir_name[0] != '/')
-	{
-	  char *path = xmalloc (PATH_MAX);
-
-	  /* FIXME: Shouldn't we use xgetcwd?  */
-#if HAVE_GETCWD
-	  if (!getcwd (path, PATH_MAX))
-	    FATAL_ERROR ((0, 0, _("Could not get current directory")));
-#else
-	  char *getwd ();
-
-	  if (!getwd (path))
-	    FATAL_ERROR ((0, 0, _("Could not get current directory: %s"),
-			  path));
-#endif
-	  chdir_name = new_name (path, chdir_name);
-	  free (path);
-	}
-    }
-
   length = string ? strlen (string) : 0;
-  name = (struct name *) xmalloc (sizeof (struct name) + length);
+  name = xmalloc (sizeof (struct name) + length);
   memset (name, 0, sizeof (struct name) + length);
-  name->next = NULL;
+  name->next = 0;
 
   if (string)
     {
       name->fake = 0;
       name->length = length;
-      /* FIXME: Possibly truncating a string, here?  Tss, tss, tss!  */
-      strncpy (name->name, string, length);
-      name->name[length] = '\0';
+      memcpy (name->name, string, length + 1);
     }
   else
     name->fake = 1;
@@ -530,7 +519,7 @@ addname (const char *string)
   name->found = 0;
   name->regexp = 0;		/* assume not a regular expression */
   name->firstch = 1;		/* assume first char is literal */
-  name->change_dir = chdir_name;
+  name->change_dir = change_dir;
   name->dir_contents = 0;
 
   if (string && is_pattern (string))
@@ -562,15 +551,13 @@ name_match (const char *path)
       struct name *cursor = namelist;
 
       if (!cursor)
-	return 1;		/* empty namelist is easy */
+	return ! files_from_option;
 
       if (cursor->fake)
 	{
-	  if (cursor->change_dir && chdir (cursor->change_dir))
-	    FATAL_ERROR ((0, errno, _("Cannot change to directory %s"),
-			  cursor->change_dir));
+	  chdir_do (cursor->change_dir);
 	  namelist = 0;
-	  return 1;
+	  return ! files_from_option;
 	}
 
       for (; cursor; cursor = cursor->next)
@@ -580,48 +567,21 @@ name_match (const char *path)
 	  if (cursor->firstch && cursor->name[0] != path[0])
 	    continue;
 
-	  /* Regular expressions (shell globbing, actually).  */
-
-	  if (cursor->regexp)
+	  if (cursor->regexp
+	      ? fnmatch (cursor->name, path, FNM_LEADING_DIR) == 0
+	      : (cursor->length <= length
+		 && (path[cursor->length] == '\0'
+		     || path[cursor->length] == '/')
+		 && memcmp (path, cursor->name, cursor->length) == 0))
 	    {
-	      if (fnmatch (cursor->name, path, FNM_LEADING_DIR) == 0)
-		{
-		  cursor->found = 1; /* remember it matched */
-		  if (starting_file_option)
-		    {
-		      free (namelist);
-		      namelist = NULL;
-		    }
-		  if (cursor->change_dir && chdir (cursor->change_dir))
-		    FATAL_ERROR ((0, errno, _("Cannot change to directory %s"),
-				  cursor->change_dir));
-
-		  /* We got a match.  */
-		  return 1;
-		}
-	      continue;
-	    }
-
-	  /* Plain Old Strings.  */
-
-	  if (cursor->length <= length
-				/* archive length >= specified */
-	      && (path[cursor->length] == '\0'
-		  || path[cursor->length] == '/')
-				/* full match on file/dirname */
-	      && strncmp (path, cursor->name, cursor->length) == 0)
-				/* name compare */
-	    {
-	      cursor->found = 1;	/* remember it matched */
+	      cursor->found = 1; /* remember it matched */
 	      if (starting_file_option)
 		{
-		  free ((void *) namelist);
+		  free (namelist);
 		  namelist = 0;
 		}
-	      if (cursor->change_dir && chdir (cursor->change_dir))
-		FATAL_ERROR ((0, errno, _("Cannot change to directory %s"),
-			      cursor->change_dir));
-
+	      chdir_do (cursor->change_dir);
+  
 	      /* We got a match.  */
 	      return 1;
 	    }
@@ -657,7 +617,8 @@ names_notfound (void)
     {
       next = cursor->next;
       if (!cursor->found && !cursor->fake)
-	ERROR ((0, 0, _("%s: Not found in archive"), cursor->name));
+	ERROR ((0, 0, _("%s: Not found in archive"),
+		quotearg_colon (cursor->name)));
 
       /* We could free the list, but the process is about to die anyway, so
 	 save some CPU time.  Amigas and other similarly broken software
@@ -668,31 +629,223 @@ names_notfound (void)
 	free (cursor);
 #endif
     }
-  namelist = (struct name *) NULL;
-  namelast = (struct name *) NULL;
+  namelist = 0;
+  namelast = 0;
 
   if (same_order_option)
     {
       char *name;
 
       while (name = name_next (1), name)
-	ERROR ((0, 0, _("%s: Not found in archive"), name));
+	ERROR ((0, 0, _("%s: Not found in archive"),
+		quotearg_colon (name)));
     }
 }
+
+/* Sorting name lists.  */
 
-/*---.
-| ?  |
-`---*/
+/* Sort linked LIST of names, of given LENGTH, using COMPARE to order
+   names.  Return the sorted list.  Apart from the type `struct name'
+   and the definition of SUCCESSOR, this is a generic list-sorting
+   function, but it's too painful to make it both generic and portable
+   in C.  */
+
+static struct name *
+merge_sort (struct name *list, int length,
+	    int (*compare) (struct name const*, struct name const*))
+{
+  struct name *first_list;
+  struct name *second_list;
+  int first_length;
+  int second_length;
+  struct name *result;
+  struct name **merge_point;
+  struct name *cursor;
+  int counter;
+
+# define SUCCESSOR(name) ((name)->next)
+
+  if (length == 1)
+    return list;
+
+  if (length == 2)
+    {
+      if ((*compare) (list, SUCCESSOR (list)) > 0)
+	{
+	  result = SUCCESSOR (list);
+	  SUCCESSOR (result) = list;
+	  SUCCESSOR (list) = 0;
+	  return result;
+	}
+      return list;
+    }
+
+  first_list = list;
+  first_length = (length + 1) / 2;
+  second_length = length / 2;
+  for (cursor = list, counter = first_length - 1;
+       counter;
+       cursor = SUCCESSOR (cursor), counter--)
+    continue;
+  second_list = SUCCESSOR (cursor);
+  SUCCESSOR (cursor) = 0;
+
+  first_list = merge_sort (first_list, first_length, compare);
+  second_list = merge_sort (second_list, second_length, compare);
+
+  merge_point = &result;
+  while (first_list && second_list)
+    if ((*compare) (first_list, second_list) < 0)
+      {
+	cursor = SUCCESSOR (first_list);
+	*merge_point = first_list;
+	merge_point = &SUCCESSOR (first_list);
+	first_list = cursor;
+      }
+    else
+      {
+	cursor = SUCCESSOR (second_list);
+	*merge_point = second_list;
+	merge_point = &SUCCESSOR (second_list);
+	second_list = cursor;
+      }
+  if (first_list)
+    *merge_point = first_list;
+  else
+    *merge_point = second_list;
+
+  return result;
+
+#undef SUCCESSOR
+}
+
+/* A comparison function for sorting names.  Put found names last;
+   break ties by string comparison.  */
+
+static int
+compare_names (struct name const *n1, struct name const *n2)
+{
+  int found_diff = n2->found - n1->found;
+  return found_diff ? found_diff : strcmp (n1->name, n2->name);
+}
+
+/* Add all the dirs in PATH, which is a directory, to the namelist.
+   If any of the files is a directory, recurse on the subdirectory.
+   CHANGE_DIR is the number of the directory that PATH is relative to.
+   DEVICE is the device not to leave, if the -l option is specified.  */
+
+static void
+add_hierarchy_to_namelist (char *path, int change_dir, dev_t device)
+{
+  char *buffer = get_directory_contents (path, device);
+
+  {
+    struct name *name;
+
+    for (name = namelist; name; name = name->next)
+      if (strcmp (name->name, path) == 0)
+	break;
+    if (name)
+      name->dir_contents = buffer ? buffer : "\0\0\0\0";
+  }
+
+  if (buffer)
+    {
+      size_t name_length = strlen (path);
+      size_t allocated_length = (name_length >= NAME_FIELD_SIZE
+				 ? name_length + NAME_FIELD_SIZE
+				 : NAME_FIELD_SIZE);
+      char *name_buffer = xmalloc (allocated_length + 1);
+				/* FIXME: + 2 above?  */
+      char *string;
+      size_t string_length;
+
+      strcpy (name_buffer, path);
+      if (name_buffer[name_length - 1] != '/')
+	{
+	  name_buffer[name_length++] = '/';
+	  name_buffer[name_length] = '\0';
+	}
+
+      for (string = buffer; *string; string += string_length + 1)
+	{
+	  string_length = strlen (string);
+	  if (*string == 'D')
+	    {
+	      if (name_length + string_length >= allocated_length)
+		{
+		  while (name_length + string_length >= allocated_length)
+		    allocated_length += NAME_FIELD_SIZE;
+		  name_buffer = xrealloc (name_buffer, allocated_length + 1);
+		}
+	      strcpy (name_buffer + name_length, string + 1);
+	      addname (name_buffer, change_dir);
+	      if (*string == 'D')
+		add_hierarchy_to_namelist (name_buffer, change_dir, device);
+	    }
+	}
+
+      free (name_buffer);
+    }
+}
+
+/* Collect all the names from argv[] (or whatever), expand them into a
+   directory tree, and sort them.  This gets only subdirectories, not
+   all files.  */
 
 void
-name_expand (void)
+collect_and_sort_names (void)
 {
+  struct name *name;
+  struct name *next_name;
+  int num_names;
+  struct stat statbuf;
+
+  name_gather ();
+
+  if (listed_incremental_option)
+    read_directory_file ();
+
+  if (!namelist)
+    addname (".", 0);
+
+  for (name = namelist; name; name = next_name)
+    {
+      next_name = name->next;
+      if (name->found || name->dir_contents)
+	continue;
+      if (name->regexp)		/* FIXME: just skip regexps for now */
+	continue;
+      chdir_do (name->change_dir);
+      if (name->fake)
+	continue;
+
+      if (deref_stat (dereference_option, name->name, &statbuf) != 0)
+	{
+	  stat_error (name->name);
+	  continue;
+	}
+      if (S_ISDIR (statbuf.st_mode))
+	{
+	  name->found = 1;
+	  add_hierarchy_to_namelist (name->name, name->change_dir,
+				     statbuf.st_dev);
+	}
+    }
+
+  num_names = 0;
+  for (name = namelist; name; name = name->next)
+    num_names++;
+  namelist = merge_sort (namelist, num_names, compare_names);
+
+  for (name = namelist; name; name = name->next)
+    name->found = 0;
 }
 
 /*-------------------------------------------------------------------------.
 | This is like name_match, except that it returns a pointer to the name it |
 | matched, and doesn't set FOUND in structure.  The caller will have to do |
-| that if it wants to.  Oh, and if the namelist is empty, it returns NULL, |
+| that if it wants to.  Oh, and if the namelist is empty, it returns null, |
 | unlike name_match, which returns TRUE.                                   |
 `-------------------------------------------------------------------------*/
 
@@ -706,7 +859,7 @@ name_scan (const char *path)
       struct name *cursor = namelist;
 
       if (!cursor)
-	return NULL;		/* empty namelist is easy */
+	return 0;
 
       for (; cursor; cursor = cursor->next)
 	{
@@ -715,24 +868,12 @@ name_scan (const char *path)
 	  if (cursor->firstch && cursor->name[0] != path[0])
 	    continue;
 
-	  /* Regular expressions.  */
-
-	  if (cursor->regexp)
-	    {
-	      if (fnmatch (cursor->name, path, FNM_LEADING_DIR) == 0)
-		return cursor;	/* we got a match */
-	      continue;
-	    }
-
-	  /* Plain Old Strings.  */
-
-	  if (cursor->length <= length
-				/* archive length >= specified */
-	      && (path[cursor->length] == '\0'
-		  || path[cursor->length] == '/')
-				/* full match on file/dirname */
-	      && strncmp (path, cursor->name, cursor->length) == 0)
-				/* name compare */
+	  if (cursor->regexp
+	      ? fnmatch (cursor->name, path, FNM_LEADING_DIR) == 0
+	      : (cursor->length <= length
+		 && (path[cursor->length] == '\0'
+		     || path[cursor->length] == '/')
+		 && memcmp (path, cursor->name, cursor->length) == 0))
 	    return cursor;	/* we got a match */
 	}
 
@@ -745,10 +886,10 @@ name_scan (const char *path)
 	{
 	  name_gather ();	/* read one more */
 	  if (namelist->found)
-	    return NULL;
+	    return 0;
 	}
       else
-	return NULL;
+	return 0;
     }
 }
 
@@ -758,25 +899,22 @@ name_scan (const char *path)
 | return all the non-found names in the namelist			 |
 `-----------------------------------------------------------------------*/
 
-struct name *gnu_list_name = NULL;
+struct name *gnu_list_name;
 
 char *
 name_from_list (void)
 {
   if (!gnu_list_name)
     gnu_list_name = namelist;
-  while (gnu_list_name && gnu_list_name->found)
+  while (gnu_list_name && (gnu_list_name->found | gnu_list_name->fake))
     gnu_list_name = gnu_list_name->next;
   if (gnu_list_name)
     {
       gnu_list_name->found = 1;
-      if (gnu_list_name->change_dir)
-	if (chdir (gnu_list_name->change_dir) < 0)
-	  FATAL_ERROR ((0, errno, _("Cannot change to directory %s"),
-			gnu_list_name->change_dir));
+      chdir_do (gnu_list_name->change_dir);
       return gnu_list_name->name;
     }
-  return NULL;
+  return 0;
 }
 
 /*---.
@@ -800,8 +938,61 @@ blank_name_list (void)
 char *
 new_name (const char *path, const char *name)
 {
-  char *buffer = (char *) xmalloc (strlen (path) + strlen (name) + 2);
+  char *buffer = xmalloc (strlen (path) + strlen (name) + 2);
 
   sprintf (buffer, "%s/%s", path, name);
   return buffer;
 }
+
+/* Return nonzero if file NAME is excluded.  Exclude a name if its
+   prefix matches a pattern that contains slashes, or if one of its
+   components matches a pattern that contains no slashes.  */
+int
+excluded_name (char const *name)
+{
+  char const *p;
+  name += FILESYSTEM_PREFIX_LEN (name);
+
+  if (excluded_filename (excluded_with_slash, name,
+			 FNM_FILE_NAME | FNM_LEADING_DIR))
+    return 1;
+
+  for (p = name; *p; p++)
+    if (((p == name || ISSLASH (p[-1])) && !ISSLASH (p[0]))
+	&& excluded_filename (excluded_without_slash, p,
+			      FNM_FILE_NAME | FNM_LEADING_DIR))
+      return 1;
+
+  return 0;
+}
+
+/* Names to avoid dumping.  */
+
+struct avoided_name
+{
+  struct avoided_name const *next;
+  char name[1];
+};
+
+static struct avoided_name *avoided_names;
+
+/* Remember to not archive NAME.  */
+void
+add_avoided_name (char const *name)
+{
+  struct avoided_name *p = xmalloc (sizeof *p + strlen (name));
+  p->next = avoided_names;
+  avoided_names = p;
+  strcpy (p->name, name);
+}
+
+/* Should NAME be avoided when archiving?  */
+int
+is_avoided_name (char const *name)
+{
+  struct avoided_name const *p;
+  for (p = avoided_names; p; p = p->next)
+    if (strcmp (p->name, name) == 0)
+      return 1;
+  return 0;
+}
