@@ -27,11 +27,12 @@
    derived from this software without specific prior written permission.
    THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
-   MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.  */
+   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  */
 
 #include "system.h"
 #include "safe-read.h"
 
+#include <getopt.h>
 #include <sys/socket.h>
 
 #ifndef EXIT_FAILURE
@@ -51,15 +52,15 @@ const char *program_name;
 static int tape = -1;
 
 /* Buffer containing transferred data, and its allocated size.  */
-static char *record_buffer = NULL;
-static size_t allocated_size = 0;
+static char *record_buffer;
+static size_t allocated_size;
 
 /* Buffer for constructing the reply.  */
 static char reply_buffer[BUFSIZ];
 
 /* Debugging tools.  */
 
-static FILE *debug_file = NULL;
+static FILE *debug_file;
 
 #define	DEBUG(File) \
   if (debug_file) fprintf(debug_file, File)
@@ -153,7 +154,7 @@ prepare_record_buffer (size_t size)
 
   record_buffer = malloc (size);
 
-  if (record_buffer == NULL)
+  if (! record_buffer)
     {
       DEBUG (_("rmtd: Cannot allocate buffer space\n"));
 
@@ -175,6 +176,113 @@ prepare_record_buffer (size_t size)
 #endif
 }
 
+/* Decode OFLAG_STRING, which represents the 2nd argument to `open'.
+   OFLAG_STRING should contain an optional integer, followed by an optional
+   symbolic representation of an open flag using only '|' to separate its
+   components (e.g. "O_WRONLY|O_CREAT|O_TRUNC").  Prefer the symbolic
+   representation if available, falling back on the numeric
+   representation, or to zero if both formats are absent.
+
+   This function should be the inverse of encode_oflag.  The numeric
+   representation is not portable from one host to another, but it is
+   for backward compatibility with old-fashioned clients that do not
+   emit symbolic open flags.  */
+
+static int
+decode_oflag (char const *oflag_string)
+{
+  char *oflag_num_end;
+  int numeric_oflag = strtol (oflag_string, &oflag_num_end, 10);
+  int symbolic_oflag = 0;
+  
+  oflag_string = oflag_num_end;
+  while (ISSPACE ((unsigned char) *oflag_string))
+    oflag_string++;
+    
+  do
+    {
+      struct name_value_pair { char const *name; int value; };
+      static struct name_value_pair const table[] =
+      {
+#ifdef O_APPEND
+	{"APPEND", O_APPEND},
+#endif
+	{"CREAT", O_CREAT},
+#ifdef O_DSYNC
+	{"DSYNC", O_DSYNC},
+#endif
+	{"EXCL", O_EXCL},
+#ifdef O_LARGEFILE
+	{"LARGEFILE", O_LARGEFILE}, /* LFS extension for opening large files */
+#endif
+#ifdef O_NOCTTY
+	{"NOCTTY", O_NOCTTY},
+#endif
+#ifdef O_NONBLOCK
+	{"NONBLOCK", O_NONBLOCK},
+#endif
+	{"RDONLY", O_RDONLY},
+	{"RDWR", O_RDWR},
+#ifdef O_RSYNC
+	{"RSYNC", O_RSYNC},
+#endif
+#ifdef O_SYNC
+	{"SYNC", O_SYNC},
+#endif
+	{"TRUNC", O_TRUNC},
+	{"WRONLY", O_WRONLY}
+      };
+      struct name_value_pair const *t;
+      size_t s;
+
+      if (*oflag_string++ != 'O' || *oflag_string++ != '_')
+	return numeric_oflag;
+
+      for (t = table;
+	   (strncmp (oflag_string, t->name, s = strlen (t->name)) != 0
+	    || (oflag_string[s]
+		&& strchr ("ABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789",
+			   oflag_string[s])));
+	   t++)
+	if (t == table + sizeof table / sizeof *table - 1)
+	  return numeric_oflag;
+
+      symbolic_oflag |= t->value;
+      oflag_string += s;
+    }
+  while (*oflag_string++ == '|');
+
+  return symbolic_oflag;
+}
+
+static struct option const long_opts[] =
+{
+  {"help", no_argument, 0, 'h'},
+  {"version", no_argument, 0, 'v'},
+  {0, 0, 0, 0}
+};
+
+static void
+usage (int status)
+{
+  if (status != EXIT_SUCCESS)
+    fprintf (stderr, _("Try `%s --help' for more information.\n"),
+	     program_name);
+  else
+    {
+      printf (_("\
+Usage: %s [OPTION]\n\
+Manipulate a tape drive, accepting commands from a remote process.\n\
+\n\
+  --version  Output version info.\n\
+  --help  Output this help.\n"),
+	      program_name);
+      fputs (_("\nReport bugs to <bug-tar@gnu.org>.\n"), stdout);
+    }
+
+  exit (status);
+}
+
 /*---.
 | ?  |
 `---*/
@@ -183,10 +291,10 @@ int
 main (int argc, char *const *argv)
 {
   char command;
-  long status;
+  ssize_t status;
 
-  /* FIXME: Localisation is meaningless, unless --help and --version are
-     locally used.  Localisation would be best accomplished by the calling
+  /* FIXME: Localization is meaningless, unless --help and --version are
+     locally used.  Localization would be best accomplished by the calling
      tar, on messages found within error packets.  */
 
   program_name = argv[0];
@@ -194,25 +302,45 @@ main (int argc, char *const *argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  /* FIXME: Implement --help and --version as for any other GNU program.  */
-
-  argc--, argv++;
-  if (argc > 0)
+  switch (getopt_long (argc, argv, "", long_opts, NULL))
     {
-      debug_file = fopen (*argv, "w");
+    default:
+      usage (EXIT_FAILURE);
+      
+    case 'h':
+      usage (EXIT_SUCCESS);
+      
+    case 'v':
+      printf ("rmt (GNU %s) %s\n%s\n%s\n", PACKAGE, VERSION,
+	      "Copyright 1999 Free Software Foundation, Inc.",
+	      _("\
+This program comes with NO WARRANTY, to the extent permitted by law.\n\
+You may redistribute it under the terms of the GNU General Public License;\n\
+see the file named COPYING for details."));
+      return EXIT_SUCCESS;
+
+    case -1:
+      break;
+    }
+
+  if (optind < argc)
+    {
+      if (optind != argc - 1)
+	usage (EXIT_FAILURE);
+      debug_file = fopen (argv[optind], "w");
       if (debug_file == 0)
 	{
 	  report_numbered_error (errno);
 	  exit (EXIT_FAILURE);
 	}
-      setbuf (debug_file, NULL);
+      setbuf (debug_file, 0);
     }
 
 top:
-  errno = 0;			/* FIXME: errno should be read-only */
+  errno = 0;
   status = 0;
   if (safe_read (STDIN_FILENO, &command, 1) != 1)
-    exit (EXIT_SUCCESS);
+    return EXIT_SUCCESS;
 
   switch (command)
     {
@@ -221,16 +349,16 @@ top:
     case 'O':
       {
 	char device_string[STRING_SIZE];
-	char mode_string[STRING_SIZE];
+	char oflag_string[STRING_SIZE];
 
 	get_string (device_string);
-	get_string (mode_string);
-	DEBUG2 ("rmtd: O %s %s\n", device_string, mode_string);
+	get_string (oflag_string);
+	DEBUG2 ("rmtd: O %s %s\n", device_string, oflag_string);
 
 	if (tape >= 0)
 	  close (tape);
 
-	tape = open (device_string, atoi (mode_string), 0666);
+	tape = open (device_string, decode_oflag (oflag_string), MODE_RW);
 	if (tape < 0)
 	  goto ioerror;
 	goto respond;
@@ -360,9 +488,9 @@ top:
 	status = safe_read (tape, record_buffer, size);
 	if (status < 0)
 	  goto ioerror;
-	sprintf (reply_buffer, "A%ld\n", status);
+	sprintf (reply_buffer, "A%ld\n", (long) status);
 	full_write (STDOUT_FILENO, reply_buffer, strlen (reply_buffer));
-	full_write (STDOUT_FILENO, record_buffer, (size_t) status);
+	full_write (STDOUT_FILENO, record_buffer, status);
 	goto top;
       }
 
@@ -435,10 +563,10 @@ top:
 
 	  if (ioctl (tape, MTIOCGET, (char *) &operation) < 0)
 	    goto ioerror;
-	  status = sizeof (operation);
-	  sprintf (reply_buffer, "A%ld\n", status);
+	  status = sizeof operation;
+	  sprintf (reply_buffer, "A%ld\n", (long) status);
 	  full_write (STDOUT_FILENO, reply_buffer, strlen (reply_buffer));
-	  full_write (STDOUT_FILENO, (char *) &operation, sizeof (operation));
+	  full_write (STDOUT_FILENO, (char *) &operation, sizeof operation);
 	}
 #endif
 	goto top;
@@ -452,9 +580,9 @@ top:
     }
 
 respond:
-  DEBUG1 ("rmtd: A %ld\n", status);
+  DEBUG1 ("rmtd: A %ld\n", (long) status);
 
-  sprintf (reply_buffer, "A%ld\n", status);
+  sprintf (reply_buffer, "A%ld\n", (long) status);
   full_write (STDOUT_FILENO, reply_buffer, strlen (reply_buffer));
   goto top;
 
